import unittest

from tempfile import mkdtemp
import shutil

import hylite
from hylite import HyScene, HyCloud, HyImage
from hylite.project import Camera, Pushbroom, blend_scenes, push_geomattr, get_blend_weights
import numpy as np
class MyTestCase(unittest.TestCase):
    def build_dummy_data(self):

        # build an example cloud
        x, y = np.meshgrid(np.linspace(-10, 10), np.linspace(-10, 10))
        xyz = np.vstack([x.ravel(), y.ravel(), np.zeros_like(x.ravel())]).T
        klm = np.zeros(xyz.shape)
        klm[:,2] = 1.0
        rgb = (np.random.rand( *xyz.shape ) * 255).astype(np.uint8)
        self.cloud = HyCloud( xyz, rgb=rgb, normals=klm )

        # build an example image
        dims = (20, 20, 3)
        self.image = HyImage( np.random.rand( *dims ) )
        self.image.set_wavelengths( hylite.RGB )

        # build associated camera
        pos = np.array([0, 0, 40])
        ori = np.array([0, 0, 90])
        fov = 25.
        self.cam = Camera( pos, ori, 'persp', fov, dims)

        # and a track
        self.swath = HyImage(np.random.rand(dims[0],100,3))
        self.swath.set_wavelengths(hylite.RGB)
        cp = np.zeros( (100, 3) )
        cp[:, 0] +=  np.linspace(-10, 10, 100)
        cp[:, 1] +=  np.linspace(-10, 10, 100)
        cp[:, 2] = 80.
        co = np.zeros( (100,3) )
        self.track = Pushbroom( cp, co, fov / dims[0], fov / dims[0], (dims[0], cp.shape[0]) )

    def test_construction(self):
        self.build_dummy_data()

        # make a test directory
        pth = mkdtemp()
        try:
            # init a scene
            S = HyScene(pth,"Scene1")

            # build using normal camera
            S.construct( self.image, self.cloud, self.cam, occ_tol=1, maxf=100, s=5 )

            # build using pushbroom camera
            S2 = HyScene(pth, "Scene2")
            S2.construct( self.swath, self.cloud, self.track, occ_tol=1, maxf=100, s=(5,1) )

            # test projections using normal camera
            cld = S.push_to_cloud( hylite.RGB, method='best' )
            img = S.push_to_image( 'klm', method='closest')
            self.assertAlmostEquals( np.nanmax(img.data), 1.0, 2)
            self.assertAlmostEquals(np.nanmax(cld.data), 1.0, 2 )

            # test projections using pushbroom camera
            cld = S2.push_to_cloud(hylite.RGB, method='best')
            img = S2.push_to_image('klm', method='closest')

            # test blending
            S.image = self.image
            S2.image = self.swath
            for method in ['equal','gsd','obliquity','distance']:
                w = get_blend_weights([S,S2],method=method,ascloud=True) # run different weighting methods
            O = blend_scenes([S,S2], w, (0,-1) )
            self.assertEqual( O.point_count(), 2500 )
            self.assertEqual(O.band_count(), 3)

            #print((O.data == 0).all() ,  np.isfinite(O.data).any() )
            #self.assertFalse( (O.data == 0).all() ) # check some data is valid [ crude, but better than nothing ]
            #self.assertTrue( np.isfinite(O.data).any() ) # check some data is valid [ crude, but better than nothing ]
        except:
            shutil.rmtree(pth)  # delete temp directory
            self.assertFalse(True, "Error - could not construct HyScene instance")

        shutil.rmtree(pth)  # delete temp directory




if __name__ == '__main__':
    unittest.main()



"""
Functions for calculating spectral angles and doing spectral angle mapping.
"""

import numpy as np

def spectral_angles(reference, spectra):
    """
    Calculate the angle between each spectra and each reference.

    Args:
        reference: a numpy array of reference spectra.
        spectra: a numpy array of spectra to compare to the references.

    Returns:
        a 2D numpy array such that [spectra i][reference j] gives the angle between the i'th spectra and j'th reference spectra.
    """

    # normalise spectra
    reference = np.array(reference) / np.linalg.norm(np.array(reference), axis=1)[:, None]
    spectra = np.array(spectra) / np.linalg.norm(spectra, axis=1)[:, None]

    # calculate angle
    out = np.zeros((len(reference), len(spectra)))
    for i in range(len(reference)):
        out[i, :] = np.arccos(np.dot(reference[i], spectra.T))

    return out


def SAM(data, ref_spec):
    """
    Apply a spectral angle classification based on reference spectra.
    Args:
        data: the HyData instance (e.g. image or cloud) to apply the classification to.
        ref_spec: a list containing lists of spectra for each class. i.e.:
                    ref_spect = [ [class1_spec1, class1_spec2],[class2_spec1, class2_spec2], ... ]
    Returns:
        a HyData instance with the same type as data containing two bands: the class index, and the spectral angle to this (closest) class.
    """
    R = []
    L = []
    for i, S in enumerate(ref_spec):
        for s in S:
            L.append(i)  # label of this spectra
            R.append(s)  # append spectra
    # calculate angles
    ang = spectral_angles(np.array(R), data.X())
    # extract classifications
    sam = np.take(L, np.argmin(ang, axis=0)).astype(np.float32)  # find best matching class
    ang = np.rad2deg(np.min(ang, axis=0)).astype(np.float32)  # calculate spectral angle

    # mask
    sam[np.isnan(data.X()).all(axis=-1)] = np.nan
    ang[np.isnan(data.X()).all(axis=-1)] = np.nan

    # reshape and return
    out = data.copy( data = False)
    out.data = np.reshape(  np.array([sam, ang]).T,  data.data.shape[:-1] + (2,) )
    out.set_wavelengths(None)
    out.set_band_names(["Class", "Angle"])
    return out


"""
Store and manipulate point clouds and hyperclouds.
"""


import numpy as np
import scipy.spatial as spatial
from scipy.spatial import KDTree
from tqdm import tqdm

import hylite
from hylite.hydata import HyData
from hylite.hyimage import HyImage
from hylite.project import proj_persp, proj_pano, rasterize, Camera



class HyCloud( HyData ):
    """
    A class for point cloud data, including (but not limited to) hyperclouds.
    """

    def __init__(self, xyz, **kwds):
        """
        Create a HyCloud from a data array.

        Args:
            xyz (ndarray): a numpy array such that data[i] = [ x, y, z ]
            **kwds: Optional keywords include:

                 - normals = a Nx3 numpy array containing point normals. Default is None.
                 - rgb = a Nx3 numpy array containing point colors. Default is None.
                 - bands = a Nxm numpy array containing additional scalar bands (e.g. hyperspectral bands). This will become the
                            data array of this cloud.
                 - band_names = a Nxm list of names corresponding to each scalar band.
                 - wavelengths = a Nxm list of hyperspectral wavelengths corresponding to each scalar band. Should be -1 for non
                                 hyperspectral bands.


        """
        assert xyz.shape[1] == 3, "Error - data array must be a list of 3D vectors."

        #pass any arguments to constructor of HyData (header file, band names)
        super().__init__( kwds.get('bands', None) , **kwds )

        self.xyz = xyz
        self.normals = kwds.get('normals', None)
        self.rgb = kwds.get('rgb', None)

        #update header stuff
        if "band names" in kwds:
            self.header['band names'] = list(kwds.get("band_names"))
        if "wavelengths" in kwds:
            self.header['wavelength'] = np.array( kwds.get("wavelengths") )
        self.header["file type"] = "Hypercloud"

    def copy(self, data=True):
        """
        Make a (deep) copy of this point cloud.

        Args:
            data (bool): True if scalar bands should be copied as well (or just geometry). Default is True. If set to false then
                  points, colors and normals are copied, but not data.
        """

        xyz = self.xyz.copy()
        rgb = None
        normals = None
        bands = None
        header = self.header.copy()
        if self.has_rgb(): rgb = self.rgb.copy()
        if self.has_normals(): normals = self.normals.copy()
        if self.has_bands() and data: bands = self.data.copy()

        #if we don't copy data, remove band names and wavelengths from header
        if not data:
            header.set_band_names(None)
            header.set_wavelengths(None)

        return HyCloud(xyz, rgb=rgb, normals=normals, bands=bands, header=header)

    ############################
    ##General cloud properties
    ############################
    def point_count(self):
        """
        Number of points in this cloud.
        """

        return self.xyz.shape[0]

    def has_rgb(self):
        """
        Do points have defined colours?
        """

        return not self.rgb is None

    def has_normals(self):
        """
        Do points have defined normals?
        """

        return not self.normals is None

    def flip_normals(self):
        """
        Flip cloud normals
        """

        if self.has_normals():
            self.normals *= -1

    def has_bands(self):
        """
        Do points have defined scalar bands?
        """

        return not self.data is None

    ############################
    ##Data management
    ############################

    def delete_bands(self):
        """
        Delete all scalar band info associated with this cloud.
        """

        self.data = None
        if "wavelength" in self.header:
            del self.header['wavelength']
        if "band names" in self.header:
            del self.header["band names"]

    def set_bands(self, data, wavelengths=None, band_names=None, copy=False):
        """
        Set the scalar bands that are associated with this point cloud. Any exisiting bands will be overwritten.

        Args:
            data (ndarray): the new scalar bands
            wavelengths (ndarray): associated wavelength data. Default is None.
            band_names (list): associated band names. Default is None.
            copy (bool): True if the data array should be copied (rather than just copying the reference). Default is False.
        """

        assert data.shape[0] == self.point_count(), "Error - scalar bands must be defined for all points."
        if copy:
            self.data = data.copy()
        else:
            self.data = data

        if not band_names is None:
            self.set_band_names(band_names)
        if not wavelengths is None:
            self.set_wavelengths(wavelengths)

    def add_bands(self, data, wavelengths=None, band_names=None, copy=False):
        """
        Append the specified bands to any existing ones.

        Args:
            data (ndarray): the new scalar bands
            wavelengths (ndarray): associated wavelength data. Default is None.
            band_names (list): associated band names. Default is None.
            copy (bool): True if the data array should be copied (rather than just copying the reference). Default is False.
        """

        assert data.shape[0] == self.point_count(), "Error - scalar bands must be defined for all points."

        if self.data is None: # no band-data exists, simply set
            self.set_bands( data, wavelengths, band_names, copy )
        else:

            if copy:
                data = data.copy()
            else:
                data = data

            self.data = np.hstack([self.data, data]).astype(self.dtype)
            if self.has_band_names():
                self.set_band_names( list(self.get_band_names()) + list(band_names))
            if self.has_wavelengths():
                self.set_wavelengths(np.hstack( [self.get_wavelengths(), wavelengths] ))

    def filter_points(self, band, val, trim=True):
        """
        Remove points based on their scalar field values.

        Args:
            band (int,float,str): index (int), wavelength (float) or name (string) of the band to filter with.
            val (int,float,tuple,bool): the filter to use. Can be of the following forms:

                        - numeric = points with this value (e.g. 0, np.nan) are deleted.
                        - tuple = all points outside (trim = True) or inside (trim=False) the range of values are deleted.
                        - boolean ndarray = all points scored as 'True' are deleted.

            trim (bool): True if points outside a tuple val should be deleted. False if points falling within the range defined by val
                 should be deleted.
        """

        b = self.get_band_index(band)

        if isinstance(val, tuple) or isinstance(val, list):
            assert len(val) == 2, "Error - filter range (val) must have length 2."
            if trim:
                msk = np.logical_or(self.data[..., b] <= val[0], self.data[..., b] >= val[1])
            else:
                msk = np.logical_and(self.data[..., b] >= val[0], self.data[..., b] <= val[1])
        elif isinstance(val, np.ndarray):
            assert val.dtype == bool, "Error - only boolean arrays can be used as point masks."
            msk = val
        else:
            msk = self.data[..., b] == val

        # convert to indices
        msk = np.argwhere(msk)

        # do deletes (n.b. we don't slice here so that the memory is actually freed)
        self.data = np.delete(self.data, msk, axis=0)
        self.xyz = np.delete(self.xyz, msk, axis=0)
        if self.has_rgb():
            self.rgb = np.delete(self.rgb, msk, axis=0)
        if self.has_normals():
            self.normals = np.delete(self.normals, msk, axis=0)

    def point_neighbourhood_operation(self, radius, function, vb=False, args=()):
        """
        Apply the specified function to each point neighbourhood (points within the specified radius).

        Args:
            radius (float): the radius around each point defining the neighbourhood.
            function: the operator to call. This should have the following form: function( radius, current_id, neighbour_ids, *args ).
            vb (bool): True if a progress bar should be created. Default is True.
            args (tuple): remaininh arguments are passed to function(...).

        Returns:
            a list of values corresponding to the returned value of function for each point in this cloud (or None if the
            function has no return value).
        """

        tree = KDTree(self.xyz, leafsize=10)  # build kdtree
        loop = range(self.xyz.shape[0])
        out = []
        if vb:
            loop = tqdm(loop, leave=False)
        for n in loop:
            # get neighbours
            N = tree.query_ball_point(self.xyz[n, :], r=radius)

            # run operation
            result = function( radius, n, N, *args )
            if result is None: # no return
                out = None
            elif out is not None:
                out.append( result )
        return np.array(out)

    def despeckle(self, radius, bands, vb=True):
        """
        Despeckle scalar fields or rgb bands associated with this point cloud using a median filter.

        Args:
            radius (float): the radius to use to define point neighbourhoods for median calculation
            bands (str,list,float,tuple): the bands to apply the median filter to. Options are: 'rgb' to smooth colours, a list of integer
                   or float band wavelengths, or a tuple of length 2.
            vb (bool): True if a progress bar should be created, as this can be a slow operation.
        """

        # build appropriate smoothing function
        if isinstance(bands, str) and 'rgb' in bands:  # smooth RGB
            def median(radius, current_id, neighbour_ids):
                self.rgb[current_id, :] = np.nanmedian(
                    np.vstack([self.rgb[neighbour_ids, :], self.rgb[[current_id], :]]), axis=0)
        elif isinstance(bands, int) or isinstance(bands, float):  # smooth single band
            idx = self.get_band_index(bands)

            def median(radius, current_id, neighbour_ids):
                self.data[current_id, idx] = np.nanmedian(
                    np.hstack([self.data[neighbour_ids, idx], self.data[[current_id], idx]]), axis=0)
        elif isinstance(bands, tuple) and len(bands) == 2:  # (min,max) tuple
            idx0 = self.get_band_index(bands[0])
            idx1 = self.get_band_index(bands[1])

            def median(radius, current_id, neighbour_ids):
                self.data[current_id, idx0:idx1] = np.nanmedian(
                    np.vstack([self.data[neighbour_ids, idx0:idx1], self.data[[current_id], idx0:idx1]]), axis=0)
        elif len(bands) > 2:  # list of bands
            bands = np.array(bands)
            if bands.dtype == 'bool':
                assert len(
                    bands) == self.band_count(), "Error - band mask has invalid shape (%s) for dataset with %d bands." % (
                    bands.shape, self.band_count())
            else:
                bands = [self.get_band_index(int(b)) for b in bands]

            def median(radius, current_id, neighbour_ids):
                subset = np.vstack([self.data[neighbour_ids, :], self.data[[current_id], :]])
                self.data[current_id, bands] = np.nanmedian(subset[:, bands], axis=0)
        else:
            assert False, 'Error - %s is an invalid band descriptor' % bands

        # apply it
        self.point_neighbourhood_operation(radius, median, vb)

    def compute_normals(self, radius, vb=True):
        """
        Compute surface normals by fitting a plane to points within  the specified distance of each point in the cloud.
        Note that this can be slow... (especially for large values of radius).

        Args:
            radius (float): the search distance for points to use in the plane fitting.
            vb (bool): True if a progress bar should be created. Default is true.
        """

        # reset normals
        self.normals = np.zeros(self.xyz.shape)

        # function for computing normal on each neighbourhood
        def cmpN( r, n, N ):
            if len(N) > 3:
                # fit plane to points
                patch = self.xyz[N, :]
                patch -= np.mean(patch, axis=0)[None, :]  # convert to barycentric coords
                u, s, vh = np.linalg.svd(patch)  # fit plane using SVD
                self.normals[n, :] = vh[2, :]

        # compute normals for each point
        self.point_neighbourhood_operation( radius, cmpN, vb=vb )

        # orient normals upwards
        self.normals[self.normals[:, 2] < 0, :] *= -1

    ############################
    ## Plotting functions
    ############################
    # noinspection PyDefaultArgument
    def render(self, cam='ortho', bands=['rgb'], **kwds ):

        """
        Renders this point cloud to a HyImage using the specified camera.

        Args:
            cam (hylite.project.Camera, str): the camera to render with. Either a Camera instance or 'ortho' to render an orthographic top-down view (default).
            step (int): the image pixel angular step (in x) for panoramic images. Default is None == square pixels.
            bands (list,str, tuple): List defining the bands to include in the output image. Elements should be one of:

                - 'rgb' = rgb
                - 'xyz' = point position
                - 'klm' = normal vectors
                - numeric = index (int), wavelength (float) of a scalar field
                - tuple of length 2 = slice of scalar fields (e.g. (0,-1) would return all bands).
                - tuple of length > 2 or list: list of band indices (int) or wavelengths (float).

                Default is ['rgb'].

            **kwds: Keyword arguments can be any of:

                 - s = the point size (in pixels). Must be an integer. Default is 1.
                 - step = skip through n points for quicker plotting (default is 1 = draw all points).
                 - fill_holes = True if 1-pixel holes should be filled. Default is False.
                 - blur = Gaussian kernel size (in pixels) to blur/smooth final image with. Default is 0 (no blur). Must be odd.
                 - despeckle = Median kernel size (in pixels) to blur/smooth final image with. Default is 0 (no blur). Must be odd.
                 - res = the pixel size (in world coordinates) used to georeference the resulting raster (if creating an orthophoto).
                         Default is one thousandth of the maximum dimension (in x or y).
                 - epsg = an epsg code used to georeference the resulting render (if creating an orthophoto). Default is 32629.
                 - depth = include the depth buffer in the output image. Default is False.

        Returns:
            a HyImage object.
        """

        # get keywords
        s = kwds.get("s",1)
        step = kwds.get("step",1)
        fill_holes = kwds.get("fill_holes", False)
        blur = kwds.get("blur", 0)
        despeckle = kwds.get("despeckle", 0)
        EPSG = kwds.get("EPSG", "EPSG:32629")
        depth = kwds.get("depth", False)

        # project points
        assert isinstance(step, int) and step >= 1, "Error - step argument must be an integer equal to or larger than 1."
        assert isinstance(EPSG, str) and "EPSG" in EPSG, "Error - EPSG must be given as string in format 'EPSG:32629' (example) "
        if isinstance(cam, str):
            # project to nadir orthophoto
            if 'ortho' in cam:
                # find world coordinate size and position
                cloudxmin = np.amin(self.xyz[::step, 0])
                cloudxmax = np.amax(self.xyz[::step, 0])
                cloudymin = np.amin(self.xyz[::step, 1])
                cloudymax = np.amax(self.xyz[::step, 1])

                # compute sensible resolution
                res = kwds.get("res", None)
                if res is None: # use default value
                    res = max(cloudxmax - cloudxmin, cloudymax - cloudymin) / 1000


                cloudxsize = int(cloudxmax - cloudxmin) / res
                cloudysize = int(cloudymax - cloudymin) / res

                # project data
                C = np.array([cloudxmin,cloudymax,0])
                pp = np.abs(self.xyz[::step,:].copy() - C[None, :])/res
                vis = np.ones(pp.shape[0], dtype=bool)
                dims=(int(cloudxsize+1/res), int(cloudysize+1/res))
            else:
                assert False, "Error - unknown camera_type. Should be 'ortho', 'perspective' or 'panorama'."
        else:
            if 'persp' in cam.proj.lower():
                dims = cam.dims
                pp, vis = proj_persp(self.xyz[::step,:], C=cam.pos, a=cam.ori,
                                     fov=cam.fov, dims=dims)
            elif 'pano' in cam.proj.lower():
                dims = cam.dims
                pp, vis = proj_pano(self.xyz[::step,:], C=cam.pos, a=cam.ori,
                                    fov=cam.fov, dims=dims, step=cam.step)
            else:
                assert False, "Error - unknown camera_type. Should be 'ortho', 'perspective' or 'panorama'."

        # gather data
        data = []
        wav = []
        nam = []

        # special case: individual band; wrap in list
        if isinstance(bands, int) or isinstance(bands, float) or isinstance(bands, str):
            bands = [bands]

        # special case: tuple of two bands; treat as slice
        elif isinstance(bands, tuple) and len(bands) == 2:
            # noinspection PyUnusedLocal
            bands = [bands]

        #loop through bands tuple/list and extract data indices/slices
        for e in bands:

            #extract based on string
            if isinstance(e, str):
                for c in e.lower():
                    if  c == 'r':
                        assert self.has_rgb(), "Error - RGB information not found."
                        data.append(self.rgb[::step,0])
                        nam.append('r')
                        wav.append( hylite.RGB[0] )
                    elif c == 'g':
                        assert self.has_rgb(), "Error - RGB information not found."
                        data.append(self.rgb[::step, 1])
                        nam.append('g')
                        wav.append( hylite.RGB[1] )
                    elif c == 'b':
                        assert self.has_rgb(), "Error - RGB information not found."
                        data.append(self.rgb[::step, 2])
                        nam.append('b')
                        wav.append( hylite.RGB[2] )
                    elif c == 'x':
                        data.append(self.xyz[::step, 0])
                        nam.append('x')
                        wav.append(-1)
                    elif c == 'y':
                        data.append(self.xyz[::step, 1])
                        nam.append('y')
                        wav.append(-1)
                    elif c == 'z':
                        data.append(self.xyz[::step, 2])
                        nam.append('z')
                        wav.append(-1)
                    elif c == 'k':
                        assert self.has_normals(), "Error - normals not found."
                        data.append(self.normals[::step, 0])
                        nam.append('k')
                        wav.append(-1)
                    elif c == 'l':
                        assert self.has_normals(), "Error - normals not found."
                        data.append(self.normals[::step, 1])
                        nam.append('l')
                        wav.append(-1)
                    elif c == 'm':
                        assert self.has_normals(), "Error - normals not found."
                        data.append(self.normals[::step, 2])
                        nam.append('m')
                        wav.append(-1)
            #extract slice
            elif isinstance(e, tuple):
                assert self.has_bands(), "Error -  band data not found."
                assert len(e) == 2, "Error - band slices must be tuples of length two."
                idx0 = self.get_band_index(e[0])
                idx1 = self.get_band_index(e[1])

                data.append(self.data[::step, idx0 : idx1])
                if self.has_band_names():
                    nam += [ self.get_band_names()[b] for b in range(idx0, idx1) ]
                else:
                    nam += [ str(b) for b in range(idx0, idx1) ]

                if self.has_wavelengths():
                    wav += [ self.get_wavelengths()[b] for b in range(idx0, idx1)]
                else:
                    wav += [ float(b) for b in range(idx0, idx1) ]
            #extract band based on index or wavelength
            elif isinstance(e,float) or isinstance(e,int):
                b = self.get_band_index( e )
                data.append(self.data[::step, b])
                if self.has_band_names():
                    nam.append(self.get_band_names()[b])
                else:
                    nam.append(str(b))
                if self.has_wavelengths():
                    wav.append(self.get_wavelengths()[b])
                else:
                    wav.append(float(b))
            else:
                assert False, "Unrecognised band descriptor %s" % b

        # rasterise and make HyImage
        assert len(data) > 0, "Error - no bands found to project?"
        _vals, _d = rasterize(pp, vis, data, dims, s)
        if depth: # add depth (z) buffer as last channel
            img = HyImage(np.dstack([_vals, _d]), header=self.header.copy() )
            img.set_band_names(nam + ['depth'])
            img.set_wavelengths(wav + [-1.0])
        else:
            img = HyImage( _vals, header=self.header.copy() )
            img.set_band_names(nam)
            img.set_wavelengths(wav)

        # if orthophoto is saved, define geotransform and project
        if isinstance(cam, str):
            if 'ortho' in cam:
                img.affine = ([cloudxmin, res, 0,  cloudymax, 0, -res])
                try:
                    img.set_projection_EPSG(EPSG)
                except:
                    #print("Warning - could not set orthoimage projection. Check GDAL installation.")
                    pass # silently continue if no GDAL available

        # postprocessing
        if fill_holes:
            img.fill_holes()
        if blur == True:
            blur = 3
        if despeckle == True:
            despeckle = 5
        if blur > 2:
            img.blur(int(blur))
        if despeckle > 2:
            img.despeckle(int(despeckle))
        return img

    def quick_plot(self, band='rgb', cam='ortho', s=1, step=1, fill_holes=False, blur=False, despeckle=False, res=None, **kwds):

        """
        Renders this point cloud using the specified camera.

        Args:
            band (str,float,int,tuple): the bands to plot. 'rgb' will plot colour, 'norm' will plot normals, 'xyz' will plot
                   coordinates. Or an index or tuple of 3-indices (mapped to rgb) can be passed to plot scalar fields.
            cam (hylite.project.Camera, str): the camera to render with. Default is 'ortho' (top down).
            s (int): point size (in pixels; must be an integer).
            step (int): skip through n points for quicker plotting (default is 1 = draw all points).
            fill_holes (bool): True if 1-pixel holes should be filled. Default is False.
            blur (bool): True if a 3x3 gaussian blur kernel is used to smooth the scene. Default is False.
            despeckle (bool): True if a 5x5 median filter should be used to denoise rendered image before plotting. Default is False.
            res (float): the resolution to plot in 'ortho' mode. Default is one thousandth of the maximum dimension (in x or y).
            **kwds: other keywords are passed to HyImage.quick_plot( ... ).

        Returns:
            Tuple containing

            - fig: the plot figure
            - ax: the plot axis
        """

        # render image
        img = self.render(cam, band, s=s, step=step, fill_holes=fill_holes, blur=blur, despeckle=despeckle, res=res)

        # plot image
        if img.band_count() >= 3:  # we have enough bands to map to rgb
            if 'rgb' in band: # edge case - rgb values should map from 0 to 1!
                kwds['vmin'] = kwds.get('vmin', 0.)
                if np.issubdtype(self.rgb.dtype, np.integer):
                    kwds['vmax'] = kwds.get('vmax', 255.)
                else:
                    kwds['vmax'] = kwds.get('vmax', 1.0)
                return img.quick_plot((0, 1, 2), **kwds)
            else:
                return img.quick_plot((0, 1, 2), **kwds)
        else:
            return img.quick_plot(0, **kwds)

    def colourise(self, bands, stretch=(1, 99)):

        """
        Map the specified bands to this clouds RGB using the specified percentile stretch.

        Args:
            bands (list, tuple): a list of 3 bands to map to r, g and b (respectively). Can be indices (integer) or wavelengths (float).
            stretch (tuple): a tuple containing the (min,max) percentiles (integer) of values (float) to use for the colour stretch. Default is (1,99).
        """

        # for greyscale mapping
        if isinstance(bands, int) or isinstance(bands, float):
            bands = [bands, bands, bands]
        if len(bands) == 1:
            bands = [bands[0], bands[0], bands[0]]
        assert len(bands) == 3, "Error - bands must have length three. No more. No less. Three."

        # convert to band indexes
        bands = list(bands)
        for i, b in enumerate(bands):
            bands[i] = self.get_band_index(b)

        self.rgb = self.data[:, bands].astype(float)

        # normalise bands
        for i, b in enumerate(bands):
            if 'data ignore value' in self.header:  # deal with data ignore values for integer arrays...
                self.rgb[self.rgb[:, i] == int(self.header['data ignore value'])] = np.nan  # make nans

            # calculate stretch values for normalisation
            minv, maxv = stretch
            if isinstance(minv, int) and isinstance(maxv, int):
                minv, maxv = np.nanpercentile(self.rgb[:, i], (minv,maxv))
            elif isinstance(minv, int):
                minv = np.nanpercentile(self.rgb[:, i], minv)
            elif isinstance(maxv, int):
                minv = np.nanpercentile(self.rgb[:, i], maxv)

            # normalise to range 0 - 1
            self.rgb[:, i] = (self.rgb[:, i] - minv) / (maxv - minv)

        # convert to uint8 (0 - 255)
        self.rgb[self.rgb < 0] = 0
        self.rgb[self.rgb > 1] = 1
        self.rgb = (self.rgb * 255.0).astype(np.uint8)

    def project(self, image, cam, bands=None, band_names=None, occ_tol=1.0, vb=True, ignore_zeros=True, trim=True):
        """
        Projects an image into this point cloud such that each a new scalar field is defined for each point.

        Args:
            image (hylite.HyImage): the image to project (as a HyImage object) or list of images.
            cam (hylite.project.Camera): the camera to project through, or a list of cameras (of the same length as image).
            bands (list,float): the bands to project (a scalar field is created for each band). Default (None) projects
                   all bands.
            band_names (list): names of the bands. Default is None (bands are given integer names).
            occ_tol (float): the occlusion tolerance (in point cloud coordinate system). Points within this distance of the z-buffer will be attributed with
                     projected data. Set to 0 to do no occlusion (i.e. all points along each camera ray will be attributed).
            vb (bool): True if progress bar should be created when projecting multiple images. Default is True.
            ignore_zeros (bool): True if zero data is not projected (but instead treated as "transparent"). Default is True.
            trim (bool) True if points that did not get any data projected onto them should be deleted from the cloud. Default is True.
        """

        # wrap in lists
        if isinstance(image, HyImage):
            image = [image]
        if isinstance(cam, Camera):
            cam = [cam]

        assert len(image) == len(cam), "Error - a camera position is required for every image."

        #calculate output data type
        if self.data is None:
            dtype = image[0].data.dtype
        else:
            dtype = self.data.dtype

        ###########################################
        # Extract header info from first image
        ###########################################
        if bands is None:
            bid = np.arange(0, image[0].band_count())  # all bands
        else:
            bid = [image[0].get_band_index(b) for b in bands]

        #extract band names from first image
        if band_names is None:
            if image[0].has_wavelengths():
                band_names = np.array(image[0].get_wavelengths())[bid]
            else:
                band_names = bid
        band_names = [str(b) for b in band_names] #ensure band names are string

        #extract wavelengths from first image
        wavelengths = None
        if image[0].has_wavelengths():
            wavelengths = np.array(image[0].get_wavelengths())[bid]

        # calculate value to use as nan
        nan = image[0].header.get('data ignore value', np.nan)

        #check size of data array.... (and warn if it's biiiiiiig)
        size = image[0].data.dtype.itemsize * self.point_count() * len(bid) / 1e9
        if size > 10:
            print("Warning - generating large hypercloud (%.1f Gb). You might run out of RAM and die?" % size, flush=True)

        # initialise data array
        if np.issubdtype( image[0].data.dtype, np.integer ):
            data = np.zeros((self.point_count(), len(bid)), dtype=np.uint32) #sum values as integer
        else:
            data = np.zeros((self.point_count(), len(bid)), dtype=np.float32) #we need floating point precision
        count = np.zeros((self.point_count(), len(bid)), dtype=np.uint32)

        # loop through images
        if vb and len(image) > 1:
            loop = tqdm(range(len(image)),leave=False)
        else:
            loop = range(len(image))
        for i in loop:

            #get camera
            _cam = cam[i]

            #get image
            _image = image[i]

            #update band wavelengths (if need be)
            #n.b. this is needed as the band indices may change if wavelengths are specified...
            if not bands is None:
                bid = [_image.get_band_index(b) for b in bands]

            # project points
            if 'persp' in _cam.proj.lower():
                pp, vis = proj_persp(self.xyz, C=_cam.pos, a=_cam.ori,
                                     fov=_cam.fov, dims=_image.data.shape)
            elif 'pano' in _cam.proj.lower():
                pp, vis = proj_pano(self.xyz, C=_cam.pos, a=_cam.ori,
                                    fov=_cam.fov, dims=_image.data.shape, step=_cam.step)
            else:
                assert False, "Error - unknown camera_type. Should be 'perspective' or 'panorama'."

            # no visible points...
            if not vis.any(): print("Warning: no visible points in image %d" % i, flush=True)

            # rasterise to make depth map
            _vals, _d = rasterize(pp, vis, self.xyz, _image.data.shape, 1)

            # mask occluded points
            if occ_tol != 0:  # occlusion is enabled
                vis[vis == True] = np.abs(
                    _d[pp[vis, 0].astype(int), pp[vis, 1].astype(int)] - pp[vis, 2]) <= occ_tol

            # extract pixels
            pixels = _image.data[pp[vis, 0].astype(int), pp[vis, 1].astype(int)][:, bid]
            valid = np.zeros(pixels.shape, dtype=count.dtype) + 1  # valid pixels

            #deal with nans/missing data
            for nanpixels in [np.isnan(pixels), pixels == nan ]: #nan == 0 for integer data...
                valid[nanpixels] = 0
                pixels[nanpixels] = 0

            #ignore zeros also?
            if ignore_zeros and nan != 0:
                valid[pixels == 0] = 0

            #accumulate
            data[vis, :] += pixels
            count[vis, :] += valid

        # calculate averages
        if np.issubdtype(data.dtype, np.integer): #use floor division
            data = data // count #data / count #np.divide(data, count[...,None], dtype = data.dtype)\
        else: #use floating point division
            data = data / count

        # store data to the scalar field array
        if not band_names is None:
            assert data.shape[-1] == len(band_names), \
                "Error - data has %d bands but %d names?" % (data.shape[-1],len(band_names))
        if not wavelengths is None:
            assert data.shape[-1] == len(wavelengths), \
                "Error - data has %d bands but %d wavelengths?" % (data.shape[-1], len(wavelengths))

        self.add_bands(data.astype(dtype), band_names, wavelengths )

        # trim
        if trim:
            self.filter_points( 0, np.sum(count,axis=-1) == 0 ) #remove zero points


import unittest
import hylite
from hylite import HyFeature
from hylite.reference.features import Minerals
import numpy as np

class MyTestCase(unittest.TestCase):
    def test_construct(self):
        assert Minerals.Mica_K is not None
        assert Minerals.Chlorite_Fe is not None

    def test_multigauss(self):
        x = np.linspace(2100., 2400., 500 )
        y = HyFeature.gaussian(x, 2200., 200., 0.5 )
        self.assertAlmostEquals( np.max(y), 1.0, 2 )
        self.assertAlmostEquals(np.min(y), 0.5, 2)

        y = HyFeature.multi_gauss(x, [2200.,2340.], [200., 200.], [0.5,0.5])
        self.assertAlmostEquals(np.max(y), 1.0, 2)
        self.assertAlmostEquals(np.min(y), 0.5, 2)

    def test_plotting(self):
        x = np.linspace(2100., 2400., 500)
        Minerals.Mica_K[0].quick_plot()

        Minerals.Mica_K[0].quick_plot(method='')

    def test_fitting(self):
        pass

if __name__ == '__main__':
    unittest.main()



"""
A list of utility functions for creating test and training datasets from labelled hyperspectral data. Note that we avoid implementing
specific supervised classification algorithms, as scikit-learn already does an excellent job of this. Hence, the following
functions are simply designed to easily extract features and labels that are compatible with scikit-learn.
"""

import numpy as np
from hylite import HyData


def get_feature_vectors(data, labels, ignore=[]):
    """
    Returns a feature vector and associated labels from a HyData instance.

    Args:
        data: the dataset (HyData instance) to extract features from.
        labels: a list of boolean point or pixel masks where True values should be associated with
                that label. Generated label indices will range from 0 to len(labels). Alternatively, if labels is
                a HyData instance (e.g. a classification image), then labels will be extacted from this.
        ignore: a list of labels to ignore (if labels is a HyData instance). E.g. [ 0 ] will ignore pixels labelled as background.
    Returns:
        A tuple containing:

        - F = a list containing a feature array for each class in labels.
        - c = a list of the number of features for each class.
    """

    # build boolean masks from HyData instance if necessary
    if isinstance(labels, HyData):
        # extract unique labels
        ll = np.unique(labels.data)

        # remove ignored labels
        for n in ignore:
            ll = np.delete(ll, np.where(ll == n))

        # sort increasing
        ll = np.sort(ll)

        # build masks
        masks = [labels.data[..., 0] == n for n in ll]

        # return features
        return get_feature_vectors(data, masks)

    # check labels do not overlap...
    assert np.max(np.sum(labels, axis=0)) == 1, "Error - class labels overlap..."

    # reshape image data
    data = data.get_raveled()

    # get features
    F = []
    c = []
    for i, lab in enumerate(labels):
        mask = lab.reshape(data.shape[0]).astype(bool)
        F.append(data[mask])
        c.append(np.sum(mask))

    return F, c


def balance( F, n=1.0):
    """
    Samples a balanced feature vector from a list of features, as returned by
    get_feature_vectors( ... ).

    Args:
        F: a list containing an array of features for each class.
        n: the number of features to extract. Default is None (extract as mean features as possible).
           If a float between 0 and 1 is passed then it is treated as a fraction of the maximum number of features.
           If an integer is passed then this number of features will be extracted (or max(counts)).

    Returns:
        A tuple containing:

         - X = a balanced feature feature vector with dimensions N_samples x M_features.
         - y = an array of length N_samples containing labels for each feature (ranging from 0 - n_classes).
    """

    c = [f.shape[0] for f in F]
    if n > 0 and n <= 1:
        n = int(n * np.min(c))
    else:
        n = int(n)
        assert n < np.min(c), "Error - unsufficient training data (%d < %d)" % (n, np.min(c))

    # balance dataset
    X = []
    y = []
    for i, f in enumerate(F):
        idx = np.random.choice(f.shape[0], n, replace=False)
        X.append(f[idx])
        y.append(np.full((n), i))

    # shuffle (just because)
    X = np.vstack(X)
    y = np.hstack(y)
    idx = np.random.choice(y.shape[0], y.shape[0], replace=False)
    return X[idx, :], y[idx]


def split(X, y, frac=0.5):
    """
    Randomly split a labeled feature set into testing and training sets.

    Args:
        X: the feature set to split.
        y: the label set to split.
        frac: the fraction of train vs test datasets. Default is 0.5 (50%).

    Returns:
        train_X, train_y, test_X, test_y = training and testing features and labels.
    """

    # extract training dataset
    n_train = int(X.shape[0] * frac)
    mask = np.full(X.shape[0], False)
    mask[np.random.choice(X.shape[0], n_train, replace=False)] = True
    return X[mask, :], y[mask], X[np.logical_not(mask), :], y[np.logical_not(mask)]


import os
import numpy as np
from pathlib import Path
import hylite.io as io
from hylite.sensors import Sensor

"""
Implementation of shared processing for all FX class sensors. This class should not be instantiated though - use FX10 and FX17 instead. 
"""
# noinspection PyAbstractClass
class FX(Sensor):

    @classmethod
    def correct_image(cls, image, verbose=False, **kwds):

        """
        Apply sensor corrections to an image. For the FX series sensors this just applies a dark correction.

        Args:
            image (hylite.HyImage): a hyImage instance of an image captured using this sensor.
            verbose (bool): true if updates/progress should be printed to the console. Default is False.
        """

        if cls.dark is None:
            print("Warning: dark calibration not found; could not perform radiance correction.")

        else:
            #convert to float
            image.data =image.data.astype(np.float32)

            #calculate pixel corrections from dark reference
            if verbose: print("Applying dark calibration... ", end="", flush="True")
            dref = np.mean(cls.dark.data, axis=1)

            #apply dark correction
            image.data -= dref[:,None,:]

            if verbose: print("DONE.")
            
        #transpose and flip (fx images are always rotated 90 deg for some reason)
        image.data = np.transpose(image.data,(1,0,2))
        image.data = np.flip(image.data,axis=1)

    @classmethod
    def correct_folder(cls, path):

        """
        Many sensors use simple/common data structures to store data/headers/dark reference etc. Hence it is often easiest
        to pass an output folder to the sensor for correction.

        Args
            path (str): a path to the folder containing the sensor specific data.

        Returns:
            A hyImage to which all sensor-specific corrections have been applied. Note that this will generally not include
            topographic or atmospheric corrections.

        """

        assert os.path.exists(path), "Error: invalid directory '$s'" % path

        assert False, "Error - folder correction for FX data is not yet implemented."

        #todo

        pass

class FX10(FX):

    """
    Specific details/methods for FX10 sensor
    """
    @classmethod
    def name(cls):
        """
        Returns this sensors name
        """
        return "FX10"

    @classmethod
    def fov(cls):
        """
        Return the (vertical) sensor field of view .
        """
        return 54.0

    @classmethod
    def ypixels(cls):
        """
        Return the number of pixels in the y-dimension.
        """
        return 1024

    @classmethod
    def xpixels(cls):
        """
        Return the number of pixels in the x-dimension (==1 for line scanners).
        """
        return 1 # line scanner

    @classmethod
    def pitch(cls):
        """
        Return the pitch of the each pixel in the y-dimension (though most pixels are square).
        """
        return 9.97 / 1000

class FX17(FX):


    """
    Specific details/methods for FX10 sensor
    """

    @classmethod
    def name(cls):
        """
        Returns this sensors name
        """
        return "FX17"

    @classmethod
    def fov(cls):
        """
        Return the (vertical) sensor field of view .
        """
        return 75.0

    @classmethod
    def ypixels(cls):
        """
        Return the number of pixels in the y-dimension.
        """
        return 640

    @classmethod
    def xpixels(cls):
        """
        Return the number of pixels in the x-dimension (==1 for line scanners).
        """
        return 1 # line scanner

    @classmethod
    def pitch(cls):
        """
        Return the pitch of the each pixel in the y-dimension (though most pixels are square).
        """
        return 18.7 / 1000

class FX50(FX):
    """
    Implementation of sensor corrections for the FX50 sensor.
    """

    @classmethod
    def name(cls):
        """
        Returns this sensors name
        """
        return "FX50"

    @classmethod
    def fov(cls):
        """
        Return the (vertical) sensor field of view .
        """
        return 45.

    @classmethod
    def ypixels(cls):
        """
        Return the number of pixels in the y-dimension.
        """
        return 640  # n.b. sensor has 384 pixels but this is resized to 401 on lens correction

    @classmethod
    def xpixels(cls):
        """
        Return the number of pixels in the x-dimension (==1 for line scanners).
        """
        return 1  # the FX50 is a line-scanner

    @classmethod
    def pitch(cls):
        """
        Return the pitch of the each pixel in the y-dimension (though most pixels are square).
        """
        return 0.07

    @classmethod
    def correct_image(cls, image, verbose=True, **kwds):

        """
        Apply sensor corrections to an image.

        Args:
            image (hylite.HyImage): a hyImage instance of an image captured using this sensor.
            verbose (bool): true if updates/progress should be printed to the console. Default is False.
            **kwds: Optional keywords include:

                 - rad = true if image should be converted to radiance by applying dark and white references. Default is True.
                 - bpr = replace bad pixels (only for raw data). Default is True.
                 - flip = true if image should be flipped (if camera mounted backwards in core
                          scanner). Default is False.

        """

        # get kwds
        rad = kwds.get("rad", True)
        bpr = kwds.get("bpr", True)

        if rad:
            if verbose: print("Converting to radiance... ", end="", flush="True")

            # convert from int to float
            image.data = image.data.astype(np.float32)

            # flag infs
            image.data[image.data == 65535.] = np.nan

            # apply dark reference
            if cls.dark is None:
                print("Warning: dark calibration not found; no dark correction was applied! Something smells dodgy...")
            else:
                dref = np.nanmean(cls.dark.data, axis=1)  # calculate dark reference
                image.data[:, :, :] -= dref[:, None, :]  # apply dark calibration

            # apply white reference (if specified)
            if not cls.white is None:
                # calculate white reference radiance
                white = cls.white.data.astype(np.float32) - dref[:,None,:]  # average each line and subtract dark reference
                refl = np.ones(white.shape[-1])  # assume white panel is pure white

                # apply white reference
                cfac = refl[None, :] / np.nanmean( white, axis=1 )
                image.data[:, :, :] *= cfac[:, None, :]

                # also estimate noise per-band (useful for eg., MNFs)
                noise = np.nanstd(white * cfac[:, None, :], axis=(0, 1))
                image.header['band noise'] = noise

            if verbose: print("DONE.")

        ###################################################################
        # replace bad (nan) pixels with an average of the surrounding ones
        ###################################################################
        if bpr:
            if verbose: print("Filtering bad pixels... ", end="", flush="True")
            invalids = np.argwhere(np.isnan(image.data) | np.isinf(image.data))  # search for bad pixels
            for px, py, band in invalids:
                n = 0
                sum = 0
                for xx in range(px - 1, px + 2):
                    for yy in range(py - 1, py + 2):
                        if xx == px and yy == py: continue  # skip invalid pixel
                        if xx < 0 or yy < 0 or xx >= image.data.shape[0] or yy >= image.data.shape[
                            1]: continue  # skip out of bounds pixels
                        if image.data[xx][yy][band] == np.nan or image.data[xx][yy][
                            band] == np.inf: continue  # maybe neighbour is nan also
                        n += 1
                        sum += image.data[xx][yy][band]
                if n > 0: sum /= n  # do averaging
                image.data[px][py][band] = sum
            if verbose: print("DONE.")

        # rotate image so that scanning direction is horizontal rather than vertical)
        image.data = np.rot90(image.data)  # np.transpose(remap, (1, 0, 2))
        image.data = np.flip(image.data, axis=1)
        image.set_band_names(None)  # delete band names as they get super annoying

    @classmethod
    def correct_folder(cls, path, **kwds):

        """
        Many sensors use simple/common data structures to store data/headers/dark reference etc. Hence it is often easiest
        to pass an output folder to the sensor for correction.

        Args:
            path (str): a path to the folder containing the sensor specific data.
            **kwds: Optional keywords include:

                - verbose = True if print outputs should be made to update progress. Default is True.
                - other keywords are passed directly to correct_image.

        Returns:
            A hyImage to which all sensor-specific corrections have been applied. Note that this will generally not include
           topographic or atmospheric corrections.

        """

        verbose = kwds.get("verbose", True)
        kwds["verbose"] = verbose

        imgs = [str(p) for p in Path(path).rglob("capture/*.hdr")]  # all image files [including data]
        dark = [str(p) for p in Path(path).rglob("capture/DARKREF*.hdr")]  # dark reference file
        white = [str(p) for p in Path(path).rglob("capture/WHITEREF*.hdr")]  # an white reference data (core scanner)
        refl = [str(p) for p in Path(path).rglob("capture/REFL*.hdr")]  # any processed reflectance data (SiSu Rock)
        for d in dark:
            del imgs[imgs.index(d)]
        for w in white:
            del imgs[imgs.index(w)]
        for r in refl:
            del imgs[imgs.index(r)]

        if len(imgs) > 1 or len(
            dark) > 1: assert False, "Error - multiple scenes found in folder. Double check file path..."
        if len(imgs) == 0 or len(
            dark) == 0: assert False, "Error - no image or dark calibration found in folder. Double check file path... %s" % path

        if verbose: print('\nLoading image %s' % imgs[0])

        # load image
        image = io.load(imgs[0])
        FX50.set_dark_ref(dark[0])

        if len(white) > 0:  # white reference exists
            FX50.set_white_ref(white[0])

        # correct
        FX50.correct_image(image, **kwds)

        # return corrected image
        return image



import unittest
import numpy as np
from tests import genCloud, genImage
from hylite.project import Camera
from hylite.project.basic import proj_persp, rasterize, proj_ortho, proj_pano
import os
from pathlib import Path
import hylite
from hylite import io
class TestHyCloud(unittest.TestCase):

    def test_projection(self):
        # load point cloud
        cloud = io.load( os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"), "hypercloud.hdr") )
        cam = cloud.header.get_camera(0)

        # project perspective
        pp, viz = proj_persp( cloud.xyz, cam.pos, cam.ori, cam.fov, cam.dims )
        self.assertTrue(viz.all())

        R,zz = rasterize( pp, viz, cloud.rgb, cam.dims, s=2 )
        self.assertTrue( np.isfinite(zz).any() )
        self.assertTrue( np.isfinite(R).any() )

        # test rendering [ or, at least run these functions... ]
        cloud.quick_plot(hylite.RGB, cam )

        rgb = cloud.rgb.copy()
        cloud.colourise( hylite.RGB, stretch=(0.0,95) )
        self.assertEqual( (rgb == cloud.rgb).all(), False ) # check that colours have changed!
        cloud.quick_plot('rgb', cam )


    def test_cloud(self):

        cloud = genCloud(npoints = 1000, nbands=10)
        self.assertEqual(cloud.point_count(), 1000)
        self.assertEqual(cloud.has_rgb(), True)
        self.assertEqual(cloud.has_normals(), True)
        self.assertEqual(cloud.has_bands(), True)

        n0 = cloud.normals[0, :].copy()
        cloud.flip_normals()
        self.assertEqual( np.sum( n0 + cloud.normals[0,:] ), 0 )

        # test generate normals
        cloud.compute_normals(1.0, vb=True)

        cloud.filter_points(0,val=(0.1,0.5),trim=True)
        self.assertGreaterEqual(np.nanmin(cloud.data[:,0]), 0.1)
        self.assertLessEqual(np.nanmax(cloud.data[:,0]), 0.5)
        self.assertLess(cloud.point_count(),1000)

        # test rendering
        cam = Camera( pos = np.array([0.0,0.0,10.0]), ori=np.array([0.0,0.0,0.0]), fov=30, proj='persp', dims=(1000,1000) )
        cloud.render( cam )

        # test projection
        image = genImage(1000,1000)
        cloud.project( image, cam)


if __name__ == '__main__':
    unittest.main()



"""
Contains measured reflectance spectra for PVC targets. Modify these or add new .txt files to change/add reference
PVC targets. For example *MyPVCSpectra.txt* would become accessible as
*hylite.reference.spectra.MyPVCSpectra*.
"""


import unittest
import numpy as np
from tests import genImage
from hylite.sensors import Fenix, FX10, Rikola_HSC2, Rikola_RSC1, Rikola
from hylite.reference.spectra import R90

class TestSensors(unittest.TestCase):
    def test_fenix(self):
        # create fake scene and dark and white reference
        scene = genImage(Fenix.ypixels(),500,450)
        dark = genImage(Fenix.ypixels(),100,450)
        white = genImage(Fenix.ypixels(),100,450)

        #give reasonable pixel magnitudes
        scene.data[:,:,0:175] *= 4095 # VNIR sensor
        scene.data[:, :, 175:] *= 65535 # SWIR sensor
        dark.data[:, :, 0:175] *= 0.1*4095  # VNIR sensor
        dark.data[:, :, 175:] *= 0.1*65535  # SWIR sensor
        white.data = (white.data + 2.0) / 3
        white.data[:, :, 0:175] *= 0.75*4095  # VNIR sensor
        white.data[:, :, 175:] *= 0.75*65535  # SWIR sensor

        Fenix.set_dark_ref( dark )
        Fenix.set_white_ref_spectra( R90 )
        Fenix.set_white_ref( white )
        Fenix.correct_image( scene, verbose=False )
        self.assertEqual( np.isfinite(scene.data).any(), True )

    def test_fx(self):
        scene = genImage(FX10.ypixels(),500,224)
        dark = genImage(FX10.ypixels(),100,224)
        white = genImage(FX10.ypixels(),100,224)

        scene.data *= 65535  # n.b. I have no idea what FX pixel numbers should go to
        dark.data *= 0.1 * 65535
        white.data = (white.data + 2.0) / 3
        white.data *= 0.75 * 65535

        FX10.set_dark_ref( dark )
        FX10.set_white_ref( white )
        FX10.correct_image(scene, verbose=False )

if __name__ == '__main__':
    unittest.main()



"""
Functions for segmenting regions of images for individual analyses. E.g., segmenting core trays or hand samples
from images taken with a core scannner.
"""

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import hylite
import cv2
from matplotlib import path
from hylite import io
from tqdm import tqdm


def readAnnotations(annot, hsi, mask_patches=False, mask_panels=False, mask_points=False, mode='average', inplace=False):
    """
    Extract pixel positions and region spectra (e.g. calibration panels) based on an annotation image.
    This image should have three RGB bands (0 - 255) and will be parsed as follows:

     - [0,0,0] = black = retain data.
     - [255,255,255] = white = mask data (replace with nan).
     - [aaa,aaa,aaa] = grays = extract regions and add averaged spectra to header file indexed by grey value.
     - [0,255,255],[255,0,255],[255,255,0]  = cyan, magenta, yellow = extract calibration panels A, B and C (respectively).
     - [n,0,0] (reds), [0,n,0] (blues), [0,0,n] (greens) = extract pixel locations and add to header file as points A, B and C (respectively). These will be ordered based on the corresponding channel value.

    Args:
        annot: the annotations image (three channels, 0-255, uint8) matching the format described above.
        hsi: a hyperspectral image with matching x- and y- dimensions to the annotation image.
        mask_patches: True if patches should be removed from output. Default is False. [ i.e. these will be retained in the image ].
        mask_panels: True if panels should be removed from output. Default is False. [ i.e. these will be retained in the image ].
        mask_points: True if points should be removed from output. Default is False. [ i.e. these will be retained in the image ].
        mode: function used to aggregate pixel values. Options are 'average' (default) or 'median'.
        inplace: True if the hsi image should be modified in-place or copied (default).

    Returns:
        a copy of the input file (if inplace is False) with the extracted points and spectra added to the header file.
    """

    # check inputs
    assert isinstance(annot, hylite.HyImage), 'Error - annot (%s) is not a HyImage instance' % type(annot)
    assert isinstance(hsi, hylite.HyImage), 'Error - hsi (%s) is not a HyImage instance' % type(hsi)
    assert annot.band_count() == 3, 'error - annot has %d bands (should = 3).' % annot.band_count()
    assert annot.xdim() == hsi.xdim(), 'error - xdim %d != %d' % (annot.xdim(), hsi.xdim())
    assert annot.ydim() == hsi.ydim(), 'error - xdim %d != %d' % (annot.ydim(), hsi.ydim())

    if 'average' in mode.lower():
        agg = np.nanmean
    elif 'median' in mode.lower():
        agg = np.nanmedian
    else:
        assert False, "Error - %s is an unknown mode. Should be 'average' or 'median'." % mode

    # create output
    if inplace:
        out = hsi
    else:
        out = hsi.copy(data=True)

    # build masks
    diff = np.diff(annot.data, axis=-1)
    retain_mask = (annot.data == 0).all(axis=-1)
    remove_mask = (annot.data == 255).all(axis=-1)
    patch_mask = (diff == 0).all(axis=-1) & np.logical_not(remove_mask) & np.logical_not(retain_mask)
    panel_mask = np.sum(annot.data != 0, axis=-1) == 2
    point_mask = np.sum(annot.data != 0, axis=-1) == 1

    # extract patches
    patches = np.unique(annot.data[patch_mask, 0])
    for i, v in enumerate(patches):
        mask = patch_mask & (annot.data[..., 0] == v)
        if mask.any():
            S = agg(hsi.data[mask, :], axis=0)
            out.header['patch %d size' % int(v)] = np.sum(mask)
            out.header['patch %d spectra' % int(v)] = S

    # extract panels
    for i, v in zip(range(3), 'abc'):
        mask = panel_mask & (annot.data[..., i] == 0)
        if mask.any():
            S = agg(hsi.data[mask, :], axis=0)
            out.header['panel %s size' % v] = np.sum(mask)
            out.header['panel %s spectra' % v] = S

    # extract points
    for i, v in zip(range(3), 'abc'):
        mask = point_mask & (annot.data[..., i] != 0)
        if mask.any():
            vals = annot.data[mask, i]
            coords = np.argwhere(mask)
            idx = np.argsort(vals)
            out.header['points %s x' % v] = coords[idx, 0]
            out.header['points %s y' % v] = coords[idx, 1]

    # apply mask
    out.data[remove_mask, :] = np.nan
    if mask_patches:
        out.data[patch_mask, : ] = np.nan
    if mask_panels:
        out.data[panel_mask, : ] = np.nan
    if mask_points:
        out.data[point_mask, : ] = np.nan

    # return
    return out

# create foreground and background masks for grab cut algorithm
def label_blocks(image, fg=None, s=8, epad=20, boost=3, erode=3, bands=hylite.RGB, vb=True, **kwds):
    """
    Segments an image into background and different hand samples based on point labels.

    Args:
        image: the image to segment.
        fg: the foreground definition (labelled with sample IDs). If None (default) then samples are pulled from the header file.
        s: the number of pixels to label at each label point (if fg is not specified).
        epad: padding around the edge of the image to be specified as background. Default is 20 pixels.
        boost: multiplication factor to exaggurate foreground background contrast.
        erode: Amount of erosion to apply to remove small regions labelled as sample. Default is 3. Set to 0 to disable.
        bands: the bands to use for segmentation. Default is hylite.RGB.
        vb: True if figures should be generated for QAQC. Default is True.
        **kwds: keywords are passed to HyImage.quick_plot( ... ) if vb is true.
    """

    # create fg array?
    if fg is None:
        fg = np.zeros((image.data.shape[0], image.data.shape[1]), dtype=np.uint64)
        points = [ image.header.get_sample_points(s) for s in image.header.get_class_names() ]
        assert len(points) > 0, "Error - image has no defined sample points. Call image.pickSamples(...) before segmenting."

        for i, fp in enumerate(points):  # enumerate samples
            for p in fp:  # enumerate points
                fg[max(p[0] - s, 0): min(p[0] + s, fg.shape[0]),
                max(p[1] - s, 0): min(p[1] + s, fg.shape[1])] = i + 1

    mask = np.full((image.data.shape[0], image.data.shape[1]), cv2.GC_PR_BGD)
    mask[0:epad, :] = cv2.GC_BGD
    mask[-epad:-1] = cv2.GC_BGD
    mask[:, 0:epad] = cv2.GC_BGD
    mask[:, -epad:-1] = cv2.GC_BGD

    # add foreground pixels
    mask[fg != 0] = cv2.GC_FGD

    # extract RGB image and boost contrast
    rgb = image.data[..., [image.get_band_index(b) for b in bands]].copy()
    rgb -= np.nanmin(rgb)
    rgb /= np.nanmax(rgb)
    rgb *= boost
    rgb[rgb > 1] = 1
    rgb = (rgb * 255).astype(np.uint8)

    # do grab cut
    bgdModel = np.zeros((1, 65), np.float64)
    fgdModel = np.zeros((1, 65), np.float64)
    mask, bgdModel, fgdModel = cv2.grabCut(rgb, mask.astype(np.uint8), None, bgdModel, fgdModel, 5,
                                           cv2.GC_INIT_WITH_MASK)
    mask = (mask == cv2.GC_PR_FGD) + (mask == cv2.GC_FGD)  # convert to boolean foreground maks

    # erode
    if erode > 0:
        kernel = np.ones((3, 3), np.uint8)
        mask = cv2.erode(mask.astype(np.uint8), kernel, iterations=1)
        mask = mask == 1

    # extract contours and create label mask
    contours,_ = cv2.findContours(mask.astype(np.uint8) * 255, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    labels = np.zeros(mask.shape, dtype=int)
    xx, yy = np.meshgrid(np.arange(mask.shape[0]), np.arange(mask.shape[1]))
    xx = xx.flatten()
    yy = yy.flatten()
    points = np.vstack([xx, yy]).T  # coordinates of each pixel
    for cnt in contours:

        # convert to matplotlib path
        verts = np.array([cnt[:, 0, 1], cnt[:, 0, 0]]).T
        verts = np.vstack([verts, verts[0][None, :]])
        if verts.shape[0] < 3:
            continue

        pth = path.Path(verts, closed=True)

        # get internal labels
        inside = pth.contains_points(points)
        inside = inside.reshape((mask.shape[1], mask.shape[0])).T
        if inside.any():
            labels[inside] = np.max(fg[inside])  # fill label

    # plot
    if vb:
        fig, ax = image.quick_plot(bands, **kwds)
        ax.imshow(np.ma.MaskedArray(fg, fg == 0).T, alpha=0.8)
        ax.imshow(np.ma.MaskedArray(labels, labels == 0).T, alpha=0.5)
        fig.show()

    # create classification file
    cls = hylite.HyImage(labels[:, :, None])
    cls.header["file type"] = "ENVI Classification"

    n = int(np.max(fg))
    cls.header['classes'] = n
    names = image.header.get_class_names()
    if names is not None and len(names) == 0:
        names = [str(i) for i in range(1, n + 1)]
        cls.header['class names'] = ['background'] + names
    cmap = mpl.cm.get_cmap('viridis')
    cls.header['class lookup'] = (np.array([cmap(i)[:3] for i in np.linspace(0, 1, n)]).ravel() * 255).astype(np.uint8)

    return cls


# do extraction
def extract_tiles(image, labels, connected=False, ids=None, erode=0):
    """
    Extract tiles that each contain a single contiguous block based on the specified classification. Useful for
    extracting individual samples from core-scanner imagery.

    Args:
        image: the hyperspectral image to extract blocks from
        labels: a HyImage instance or numpy array with different samples labelled as non-0 values.
        connected: True if separate groups of pixels with the same label should be split
                   into different tiles using cv2.connectedComponents. Default is True.
        ids: a list of label ids to extract. If None (default), all non-zero classes will
             be extracted.
        erode: number of pixels to erode from mask before extracting tiles. Default is 0.
    """

    # get labels array
    if isinstance(labels, hylite.HyImage):
        labels = labels.data[..., 0]

    # get ids
    if ids is None:
        ids = np.unique(labels)
        ids = ids[ids != 0]  # remove zero

    # extract tiles
    tiles = []
    sampleID = []
    for i in ids:  # loop through labels

        # get mask
        mask = labels == i
        if not mask.any():
            continue  # can be the case if erode removed entire labels

        if erode > 0:
            mask = cv2.erode(mask.astype(np.uint8), np.ones((erode, erode,))) == 1

        if not connected:
            # easy - don't bother to separate
            pixels = np.argwhere(mask)  # get valid pixels
            local = pixels - np.min(pixels, axis=0)[None, :]  # translate to get pixels in tile

            tile = hylite.HyImage(np.full((np.max(local[:, 0] + 1),
                                           np.max(local[:, 1] + 1), image.band_count()), np.nan))
            tile[local[:, 0], local[:, 1], :] = image.data[pixels[:, 0], pixels[:, 1], :]
            tile.set_wavelengths(image.get_wavelengths())

            # store label and tile
            sampleID.append(i)
            tiles.append(tile)
        else:
            # check connected components first
            nc, cl = cv2.connectedComponents(mask.astype(np.uint8))
            for n in range(nc):  # loop through components
                comp_mask = cl == n
                if not mask[comp_mask].all():
                    continue  # this component is background

                mask2 = mask & comp_mask

                # extact pixels
                pixels = np.argwhere(mask2)  # get valid pixels
                local = pixels - np.min(pixels, axis=0)[None, :]  # translate to get pixels in tile

                tile = hylite.HyImage(np.full((np.max(local[:, 0] + 1),
                                               np.max(local[:, 1] + 1), image.band_count()), np.nan))
                tile[local[:, 0], local[:, 1], :] = image.data[pixels[:, 0], pixels[:, 1], :]
                tile.set_wavelengths(image.get_wavelengths())

                # store label and tile
                sampleID.append(i)
                tiles.append(tile)

    return tiles, sampleID

def group_tiles( tiles, groups, ids=None, rotate=True, ignore=[] ):
    """
    Group a list of image tiles based on corresponding group IDs.

    Args:
        tiles: a list of HyImage tiles.
        groups: a list of integers corresponding to the grouping to arrange tiles by.
        ids: a list of sample IDs to create an id classification (or None).
        rotate: True if tiles should be rotated such that they're longer dimension aligns with the x-axis.
        ignore: a list of group id's that should be ignored/excluded from this grouping.
    Returns:
        A tuple containing:

         - tiled_image = a HyImage instance containing the arranged tiles.
         - tiled_label = a HyImage classification containing the class indices used to arrange the tiles.
         - [tiled_id] = a HyImage classification containing individual sample IDs (if ids is not None).
         - bounds = the bounds (x,y,width,height) of each group of tiles. Useful for labelling or separation.
     """

    # rotate tiles
    if rotate:
        for t in tiles:
            if t.xdim() < t.ydim():
                t.rot90()

    # calculate output image dimensions
    widths = []
    heights = []

    groups = np.array(groups)
    tiles = np.array(tiles)
    for t in np.unique(groups):
        if t in ignore:
            continue # skip ignored groups
        samples = np.argwhere(groups == t)[:, 0]
        widths.append(np.sum([t.xdim()+1 for t in tiles[samples]]))
        heights.append(np.max([t.ydim()+1 for t in tiles[samples]]))

    # create output arrays
    data = np.full((np.max(widths), np.sum(heights), tiles[0].band_count()), np.nan, dtype=tiles[0].data.dtype)
    label = np.full((np.max(widths), np.sum(heights), 1), 0, dtype=np.uint16)
    if ids is not None:
        ids = np.array(ids, dtype=np.uint16)
        sid = np.full((np.max(widths), np.sum(heights), 1), 0, dtype=np.uint16)

    # transfer data and labels
    _yoffs = 0
    bounds = []
    names = []
    for n in np.unique(groups):
        if n in ignore:
            continue # skip ignored groups
        samples = np.argwhere(groups == n)[:, 0]
        _xoffs = 0
        bb = [_xoffs,_yoffs]
        for i, _t,in enumerate(tiles[samples]):
            data[_xoffs:(_xoffs + _t.xdim()), _yoffs:(_yoffs + _t.ydim()), :] = _t.data
            label[_xoffs:(_xoffs + _t.xdim()), _yoffs:(_yoffs + _t.ydim()), 0] = n + 1
            if not ids is None:
                sid[_xoffs:(_xoffs + _t.xdim()), _yoffs:(_yoffs + _t.ydim()), 0] = ids[samples][i]

            _xoffs += _t.xdim()+1
        _yoffs += np.max([t.ydim() for t in tiles[samples]])+1
        bb += [_xoffs - bb[0], _yoffs - bb[1]]
        bounds.append(bb)
        names.append(n)

    # label areas with no data as background
    label[np.logical_not(np.isfinite(data).any(axis=-1))] = 0
    if ids is not None:
        sid[np.logical_not(np.isfinite(data).any(axis=-1))] = 0

    # make output images
    tiled_image = hylite.HyImage(data, header = tiles[0].header.copy())

    tiled_label = hylite.HyImage(label)
    tiled_label.set_band_names(["Class"])
    tiled_label.header['file type'] = 'ENVI Classification'
    tiled_label.header['class names'] = ['Background'] + ["Class %d" % (g+1) for g in names]
    tiled_label.header['classes'] = len(names) + 1

    if ids is not None:
        tiled_id = hylite.HyImage(sid)
        tiled_id.set_band_names(["ID"])
        tiled_id.header['file type'] = 'ENVI Classification'
        tiled_id.header['class names'] = ['Background'] + [str(i) for i in range(1, np.max(ids))]
        tiled_id.header['classes'] = np.max(ids) + 1

        return tiled_image, tiled_label, tiled_id, bounds

    return tiled_image, tiled_label, bounds

def build_core_template(images, N=5, thresh=40, vb=True):
    """
    Overlay images of core trays from e.g. a drillhole to calculate a template that is used for extracting
    individual core segments and is robust to data quirks (e.g. empty trays). All images must be identical
    dimensions and properly co-aligned.

    Args:
        images: a list of co-aligned images of different core trays build template with.
        N: the number of cores per tray. Default is 5.
        thresh: percentile used to separate foreground from background. Default is 40. Higher values ensure
                proper separation of cores, but will crop data more closely.
        vb: True if a tqdm progress bar should be printed.
    """

    # sum valid pixels
    valid = None
    loop = images
    if vb:
        loop = tqdm(images, leave=False, desc="Building template")
    for i in loop:
        if isinstance(i, str):  # load image if need be
            i = io.load(i)
        if valid is None:  # init valid if need be
            valid = np.zeros(i.data.shape[:-1])
        if not i is None:
            valid += np.isfinite(i.data).all(axis=-1)  # accumulate valid pixels

    # do threshold
    valid = valid > np.percentile(valid, thresh)
    if valid.shape[1] > valid.shape[0]:
        valid = valid.T

    # label components
    num_labels, labels_im = cv2.connectedComponents((valid.T > np.percentile(valid, 40)).astype(np.uint8))

    # take top N labels by area.
    area = [np.sum(labels_im == i) for i in range(num_labels)]
    thresh = np.sort(area)[::-1][N]
    l = 1
    for i in range(1, num_labels):
        if area[i] >= thresh:
            labels_im[labels_im == i] = l
            l += 1
        else:
            labels_im[labels_im == i] = 0  # background

    # return
    return hylite.HyImage(labels_im.T)

def unwrap_core(image, template, stack='linear', **kwds):
    """
    Unwrap drillholes in a core tray and stack them end to end along the x-axis, with the top at the left. Note that
    the top-left of the core tray is considered to be the "top", so the image may need to be flipped/mirrored prior
    to calling this function.

    Args:
        image: the image to unwrap. Background (non-core pixels) must have already been masked and set as np.nan.
        template: A labelled template
        stack: 'line' or 'separate'; either align cores in a single stick ('line'),
                or return individual tiles ('separate').
        **kwds: Keywords can include:

             - start = the meterage of the top of the core. Will be added to the image header.
             - end = the meterage of the bottom of the core. Will be added to the image header.
             - id = the id of this core tray.
    """

    # extract tiles
    tiles, ids = extract_tiles(image, template)

    # sort tiles into order
    idx = np.argsort(ids)
    tiles = np.array(tiles)[idx]

    # stack
    N = np.max(template.data)
    if 'sep' in stack.lower():

        # attribute tiles with position
        if 'start' in kwds and 'end' in kwds:
            l = (kwds['end'] - kwds['start']) / N  # length of a single stick of core
            start = np.linspace(kwds['start'], kwds['end'] - l, N)
            for i, t in enumerate(tiles):
                t.header['core start'] = start[i]  # start of each stick of core
                t.header['core end'] = start[i] + l  # end of each stick of core
        if 'id' in kwds:
            for i, t in enumerate(tiles):
                t.header['core id'] = kwds['id']

        # return
        return list(tiles)

    elif 'lin' in stack.lower():

        # attribute output with position
        out = image.copy()
        if 'start' in kwds:
            out.header['core start'] = kwds['start']
        if 'end' in kwds:
            out.header['core end'] = kwds['end']
        if 'id' in kwds:
            out.header['core id'] = kwds['id']

        # calculate width
        width = int(np.max([t.ydim() for t in tiles]))
        length = np.sum([t.xdim() for t in tiles])

        # reshape image
        data = np.full((length, width, image.band_count()), np.nan)
        x0 = 0
        for i, t in enumerate(tiles):
            data[x0:(x0 + t.xdim()), 0:t.ydim(), :] = t.data
            x0 += t.xdim()

        out.data = data
        return out
    else:
        assert False, "Error - %s is an unknown stacking." % stack

def tray_to_stick(tray, N):
    """
    Break a unwrapped tray into individual sticks.

    Args:
        tray: a HyImage instance containing a linearly unwrapped tray.
        N: the number of chunks of core to break it into.
    """

    # check orientation
    if tray.ydim() > tray.xdim():
        data = np.transpose(tray.data, (1, 0, 2))
    else:
        data = tray.data
    assert data.shape[0] >= data.shape[1], "Error - invalid shape? %s" % str(tray.data.shape)

    # slice
    step = int(data.shape[0] / N)
    slices = [data[(step * i):(step * (i + 1)), :, :] for i in range(N)]

    # build images
    start = float(tray.header['core start'])
    end = float(tray.header['core end'])
    zz = np.linspace(start, end, N + 1)
    images = []
    for i in range(N):
        image = hylite.HyImage(slices[i], header=tray.header.copy())
        image.header['core start'] = zz[i]
        image.header['core end'] = zz[i + 1]
        images.append(image)
    return images

def composite_cores(trays, pad=2):
    """
    Composites a list of uwrapped core trays (cf. unwrap_core(...)) into a single image such that each column
    represents a single tray (arranged vertically).

    Args:
        trays: a list of HyImage instances containing unwrapped core trays. The header of these
               images must contain localisation information; specifically the "core start" and "core end"
               keywords.
        pad: the padding (in pixels) between core segments.
    """

    # calculate order and identify breaks in core
    start = np.array([t.header['core start'] for t in trays])
    end = np.array([t.header['core end'] for t in trays])
    idx = np.argsort(start)
    breaks = np.argwhere(end[idx][:-1] - start[idx][1:] != 0)

    # rotate cores if need be
    for i, t in enumerate(trays):
        if t.xdim() > t.ydim():
            t = t.copy()
            t.data = np.transpose(t.data, (1, 0, 2))
            trays[i] = t

    # calculate dims and create output
    width = np.max([t.xdim() for t in trays]) + pad
    height = np.max([t.ydim() for t in trays])
    data = np.full((width * len(trays), height, trays[0].band_count()), np.nan)

    # transfer data
    x0 = 0
    y0 = 0
    px = [0]
    for i, n in enumerate(idx):
        core = trays[n].data
        x1 = x0 + core.shape[0]
        y1 = core.shape[1]
        data[x0:x1, y0:y1, :] = core
        x0 += width
        px.append(x0)

    # create output image
    out = hylite.HyImage(data)
    out.set_wavelengths(trays[0].get_wavelengths())
    out.header['wavelength units'] = trays[0].header.get('wavelength units', 'nm')
    out.header['core starts'] = start[idx]
    out.header['core ends'] = end[idx]
    out.header['core width'] = width  # width, in pixels, of individual core
    if len(breaks) > 0:
        out.header['core breaks'] = breaks

    return out

def plot_drillhole(composite, N=5, maxN=25, **kwds):
    """
    Plot a composite drillhole as created by composite_core. This plots the image data, but includes annotations
    the specify the distances along/down hole.

    Args:
        N: the number of sticks per tray. Used to ensure start/end labels are in the correct place.
        max_trays: the maximum number of trays per row. Default is 25.
    *Keywords*: are passed to HyImage.quick_plot( ... )
    """

    assert 'core width' in composite.header, "Error - no core width information in header file."
    assert 'core starts' in composite.header, "Error - no core position information in header file."
    assert 'core ends' in composite.header, "Error - no core position information in header file."

    # get data on dimensions and position
    width = int(composite.header['core width'])
    starts = composite.header['core starts']
    ends = composite.header['core ends']
    trayWidth = width * N  # width of one tray
    nTrays = int(composite.xdim() / trayWidth)

    if isinstance(starts, str):
        starts = np.fromstring(starts, sep=',')
    if isinstance(ends, str):
        ends = np.fromstring(ends, sep=',')

    # break into subimages
    nrows = int(np.ceil(nTrays / maxN))
    row_images = []
    for i in range(nrows):
        # copy data subset
        data = np.full((trayWidth * maxN, composite.ydim(), composite.band_count()), np.nan)
        x0 = i * trayWidth * maxN
        x1 = min((i + 1) * trayWidth * maxN, composite.xdim())
        data[0:(x1 - x0), :, :] = composite.data[x0:x1, :, :]

        # make new hyimage and store
        img = hylite.HyImage(data)
        if composite.has_wavelengths():
            img.set_wavelengths(composite.get_wavelengths())
        row_images.append(img)

    # make axes and plot sub-images
    dims = (18, nrows * 18 * row_images[0].ydim() / row_images[0].xdim())
    fig, ax = plt.subplots(nrows, 1, figsize=dims)
    ax = np.array(ax).ravel() # ensure ax is an array
    for i, a, im in zip(range(nrows), ax, row_images):
        im.quick_plot(ax=a, **kwds)
        a.set_yticks([])

        # set xticks
        _ends = np.array(ends[(N - 1 + i * maxN * N): min(N - 1 + (i + 1) * N * maxN, len(ends))][::N])
        _starts = np.array(starts[(i * maxN * N): min((i + 1) * N * maxN, len(starts))][::N])
        breaks = np.argwhere(_starts[1:] != _ends[:-1])  # calculate breaks where core trays don't line up

        # set ticks
        xticks = np.linspace(trayWidth, (len(_ends)) * trayWidth, len(_ends))
        a.set_xticks(xticks)
        a.set_xticklabels(["%.2f" % l for l in _ends])

        # add breaks
        # if len(breaks) > 0:
        #    for b in breaks[0]:
        #        a.axvline(xticks[b],color='r')

    fig.tight_layout()
    return fig, ax

def map_depth(image):
    # check we have the necessary info
    assert 'core width' in image.header, "Error - no core width information in header file."
    assert 'core starts' in image.header, "Error - no core position information in header file."
    assert 'core ends' in image.header, "Error - no core position information in header file."

    # get data on dimensions and position
    width = int(image.header['core width'])
    starts = image.header['core starts']
    ends = image.header['core ends']

    if isinstance(starts, str):
        starts = np.fromstring(starts, sep=',')
    if isinstance(ends, str):
        ends = np.fromstring(ends, sep=',')

    # build distance from top of each stick of core
    zz = np.zeros((image.data.shape[0], image.data.shape[1]))
    zz += np.linspace(0, 1, image.data.shape[1]) * (ends[0] - starts[0])

    # add value from previous core
    for i, x in enumerate(range(0, image.data.shape[0], width)):
        zz[x:min(x + width, image.data.shape[0])] += starts[i]

    # mask
    zz[np.logical_not(np.isfinite(image.data).any(axis=-1))] = np.nan

    return zz


"""
Implementations basic unsupervised classification algorithms, largely based on scipy.
"""
import numpy as np
from hylite import HyData
from hylite.analyse.sam import spectral_angles
from scipy.cluster.hierarchy import average, dendrogram, fcluster
import matplotlib.pyplot as plt
from hylite.filter import PCA

def cluster_hierarchical( data, nclasses, distance='SAM', vb=False, labels=None, **kwds ):
    """
    Clusters a HyData instance using a hierarchical clustering algorithm and the
    specified distance metric. Note that this can be very slow for large datasets, but
    can work well for spectral libraries or small images.

    Args:
        data: a HyData instance to cluster.
        nclasses: the number of classes to extract (level to slice the tree at).
        distance: the distance metric to use. Options are 'SAM' (spectral angle) and 'pca' (use top n principal components).
        vb: True if a preview plot of the dendrogram should be created.
        labels: labels for the dendrogram, or None (default).
        nbands: the number of mnf bands to use if distance = MNF. Default is 10.

    Returns:
        A tuple containing:

         - labels = an integer array of class labels.
         - Z = the linkage matrix returned by scipy.
    """

    assert isinstance( data, HyData ), "Error - dataset is not a HyData instance."


    # calculate distance matrix
    if 'sam' in distance.lower(): # calculate spectral angle distance matrix
        X = data.get_raveled()
        M = spectral_angles( X, X )
    elif 'pca' in distance.lower(): # calculate euclidian distance between PCA bands
        # calculate PCA
        pca = PCA(data, kwds.get("pca", 10))
        X = pca.get_raveled()

    else:
        assert False, "%s is an unknown distance type." % distance

    # do clustering
    Z = average(M)
    if vb:
        dendrogram(Z, labels=labels)
        plt.show()

    # cut tree
    C = fcluster( Z, nclasses, criterion='maxclust')

    if data.is_image():
        C = C.reshape( data.xdim(), data.ydim() )

    return C, Z


"""
Perspective, panoramic and orthographic projections.
"""


import numpy as np
from scipy import spatial
from numba import jit, prange
import cv2

def proj_persp( xyz, C, a, fov, dims, normals=None):
    """
    Project 3d point xyz based on a pinhole camera model.

    Args:
        xyz (ndarray): a nx3 numpy array with a position vector (x,y,z) in each column.
        C (ndarray): the position of the camera.
        a (ndarray): the three camera rotatation euler angles (appled around x, then y, then z == pitch, roll, yaw). In DEGREES.
        fov (float): the vertical field of view.
        dims (tuple): the image dimensions (width, height)
        normals (ndarray): per-point normals. Used to do backface culling if specified. Default is None (not used).

    Returns:
        A tuple containing:

        - pp = a (n,3) array containg the project point coordinates (x,y) and distance from the camera (z).
        - vis = a (n,) array scored with True if the corresponding point is visible from the camera.
    """

    #transform origin to camera position
    xyz = xyz - C[None,:]

    # backface culling
    vis = np.full(xyz.shape[0],True,dtype=bool)
    if not normals is None:
        ndv = np.einsum('ij, ij->i', normals, xyz) #calculate dot product between each normal and vector from camera to point
                                                   #(which is the point position vector in this coordinate system)
        vis = ndv < 0 #normal is pointing towards the camera (in opposite direction to view)

    #apply camera rotation to get to coordinate system
    #  where y is camera up and z distance along the view axis
    R = spatial.transform.Rotation.from_euler('XYZ',-a,degrees=True).as_matrix()
    #xyz = np.dot(xyz, R)
    xyz = xyz@R

    #calculate image plane width/height in projected coords
    h = 2 * np.tan( np.deg2rad( fov / 2 ) )
    w = h * dims[0] / float(dims[1])

    #do project and re-map to image coordinates such that (0,0) = lower left, (1,1) = top right)
    px = ((w/2.0) + (xyz[:,0] / -xyz[:,2])) / w
    py = ((h/2.0) + (xyz[:,1] / xyz[:,2])) / h
    pz = -xyz[:,2] #return positive depths for convenience

    #calculate mask showing 'visible' points based on image dimensions
    vis = vis & (pz > 0) & (px > 0) & (px < 1) & (py > 0) & (py < 1)

    #convert to pixels
    px *= dims[0]
    py *= dims[1]

    return np.array([px,py,pz]).T, vis

def proj_pano(xyz, C, a, fov, dims, step=None, normals=None):
    """
    Project 3d point xyz based on a pinhole camera model.

    Args:
        xyz (ndarray): a nx3 numpy array with a position vector (x,y,z) in each column.
        C (ndarray): the position of the camera.
        a (ndarray): the three camera rotatation euler angles (appled around x, then y, then z == pitch, roll, yaw). In DEGREES.
        fov (float): the vertical field of view.
        dims (tuple): the image dimensions (width, height)
        step (float): the angular step (degrees) between pixels in x.  If None, this is calculated to ensure square pixels.
        normals (ndarray): per-point normals. Used to do backface culling if specified. Default is None (not used).

    Returns:
        A tuple containing:

        - pp = a (n,3) array containg the project point coordinates (x,y) and distance from the camera (z).
        - vis = a (n,) array scored with True if the corresponding point is visible from the camera.
    """

    # transform origin to camera position
    xyz = xyz - C[None, :]

    # backface culling
    vis = np.full(xyz.shape[0], True, dtype=bool)
    if not normals is None:
        ndv = np.einsum('ij, ij->i', normals,
                        xyz)  # calculate dot product between each normal and vector from camera to point
        # (which is the point position vector in this coordinate system)
        vis = ndv < 0  # normal is pointing towards the camera (in opposite direction to view)

    # apply camera rotation to get to coordinate system
    #  where y is camera up and z distance along the view axis
    R = spatial.transform.Rotation.from_euler('XYZ', -a, degrees=True).as_matrix()
    #xyz = np.dot(xyz, R)
    xyz = xyz@R

    # calculate image height (in projected coords) for vertical perspective project
    h = 2 * np.tan(np.deg2rad(fov / 2))

    # calculate position in spherical coords
    pz = np.sqrt(xyz[:, 0] ** 2 + xyz[:, 2] ** 2)  # distance from axis of rotation (world units)
    py = ((h / 2.0) + (xyz[:, 1] / -pz)) / h  # perspective project vertically (0 - 1; pin-hole camera model)
    px = np.arctan2(xyz[:, 0], -xyz[:, 2])  # rotation left/right of view axis (in radians)

    # convert x to pixels (slightly tricky)
    if step is None:
        step = fov / dims[1]  # angular pixel size in y = angular pixel size in x, assuming pixels are square...
    px = px / np.deg2rad(step) + (dims[0] / 2)

    # convert y to pixels (easy)
    py *= dims[1]

    # calculate mask showing 'visible' points based on image dimensions
    vis = vis & (px > 0) & (px < dims[0]) & (py > 0) & (py < dims[1])

    return np.array([px, py, pz]).T, vis

def proj_ortho( xyz, C, V, s=1.0 ):
    """
    Project points onto a plane (orthographic projection).

    Args:
        xyz (ndarray): a nx3 numpy array with a position vector (x,y,z) in each column.
        C (ndarray): the position of the viewing plane origin (will become pixel 0,0).
        V (ndarray): the [x,y,z] viewing/projection vector (normal to the projection plane).
        s (float): the scale factor to transform world coordinates into image coordinates. Default is 1.0 (keep world coords).

    Returns:
        A tuple containing:

         - an ndarray with projected point coordinates (x,y,depth)
         - an array containing True if each point is visible and False for points behind the viewing plane.
    """

    xyz = xyz - C[None, :]  # center origin on camera
    #pz = np.dot(xyz, V)  # calculate depths (distances from plane)
    pz = xyz@V
    xyz -= (V[None, :] * pz[:, None])  # project onto plane (by removing depth)
    return s*np.array([ xyz[:,0], xyz[:,1], pz]).T, pz > 0

@jit(nopython=True)
def _rast( points, s, dims, vals  ):
    """
    Fast loop function for rasterise.
    """
    # create image arrays
    out = np.full((dims[0], dims[1], vals.shape[1]), np.nan)
    depth = np.full((dims[0], dims[1]), np.inf)

    for i in range(points.shape[0]):
        x = int(points[i,0]) # cast point to integer
        y = int(points[i,1])
        z = points[i,2]

        # double check point in the image
        if x < 0 or x >= dims[0] or y < 0 or y > dims[1]:
            continue  # skip

        if z < depth[x, y]:  # in front of depth buffer?
            for dx in np.arange(-s+1,s):
                for dy in np.arange(-s+1,s):
                    _x = x+dx
                    _y = y+dy
                    if _x < 0: continue # skip invalid
                    if _y < 0: continue
                    if _x >= dims[0]: continue
                    if _y >= dims[1]: continue
                    out[ _x, _y, : ] = vals[i]
                    depth[ _x, _y ] = z

    return out, depth


def rasterize(points, vis, vals, dims, s=1):
    """
    Rasterizes projected points onto an image grid.

    Args:
        points (ndarray): the points (px,py,depth) to rasterise,
                as returned by proj_persp or proj_pano.
        vis (ndarray): a boolean array that is True if the points are visible (as returned by
                proj_persp or proj_pano).
        vals (ndarray): Nxm array of (m) additional point values (e.g. position, normals, colour) to add to rasterize. Alternatively
              a list of Nxm numpy arrays can be passed (these will be concatentated).
        dims (ndarray): the size of the image grid.
        s (int): the size of the points in pixels. Default is 1.

    Returns:
        A tuple containing:

        - raster = rasterised image with m bands corresponding to vals.
        - depth = the depth buffer.
    """

    #vals is a list of numpy arrays - stack them
    if isinstance(vals, list):
        if len(vals[0].shape) == 1: #list of 1-d arrays; need to v-stack
            vals = np.vstack(vals).T
        else:                       #list of n-d arrays; need to h-stack
            vals = np.hstack(vals)

    #shoulder case - vals is a 1D numpy array, so we need to add second axis
    if len(np.array(vals).shape) == 1:
        vals = vals[None, :].T

    # cull invisible points
    points = np.array(points[vis, :], dtype=float)
    vals = np.array(vals[vis, :], dtype=float)

    # loop through points
    assert points.shape[1] == 3, "Error - points array should have shape (N,3)."
    assert isinstance(s, int), "Error - size (s) must be an integer."
    if s < 1: s = 1 # s must be at least 1.
    return _rast(points, int(s), dims, vals ) # call compiled loop with numba

    #for i, p in enumerate(points):
    #    x = int(p[0])
    #    y = int(p[1])
    #    z = p[2]
    #
    #    # double check point in the image
    #    if x < 0 or x >= dims[0] or y < 0 or y > dims[1]:
    #        continue  # skip
    #
    #    if z < depth[x, y]:  # in front of depth buffer?
    #        out[(x - s):(x + s + 1), (y - s):(y + s + 1), :] = vals[None, None, i]
    #        depth[(x - s):(x + s + 1), (y - s):(y + s + 1)] = z


def pix_to_ray_persp(x, y, fov, dims):
    """
    Transform pixel coordinates to a unit direction (ray) in camera coordinates using a
    perspective pinhole camera model.

    Args:
        x (int): the pixel x-coordinate. Cannot be array (only works for single pixels).
        y (int): the pixel y-coordinate. Cannot be array (only works for single pixels).
        fov (float): the camera's vertical field of view (degrees)
        dims (tuple): the dimensions of the image the pixels are contained in.

    Returns:
        A numpy array with the components of the light ray in camera coordinates (z == view direction, y == up)
    """

    aspx = dims[0] / float(dims[1])
    h = 2 * np.tan(np.deg2rad(fov / 2))
    Px = (2 * (x / dims[0]) - 1) * h / 2 * aspx
    Py = (2 * (y / dims[1]) - 1) * h / 2
    ray = np.array([Px, -Py, -1])
    return ray / np.linalg.norm(ray)  # return normalized ray

def pix_to_ray_pano(x, y, fov, step, dims):
    """
    Transform pixel coordinates to a unit direction (ray) in camera coordinates using a
    panoramic camera model.

    Args:
        px: the pixel x-coordinate. Cannot be array (only works for single pixels).
        py: the pixel y-coordinate. Cannot be array (only works for single pixels).
        fov: the camera's vertical field of view (degrees)
        step: the angular step between pixels in the x-direction.
        dims: the dimensions of the image the pixels are contained in.

    Returns:
        numpy array with the components of the light ray in camera coordinates (z == view direction, y == up)
    """

    # calculate image plane properties
    h = 2 * np.tan(np.deg2rad(fov / 2))

    # calculate ray in y-z plane
    Py = (2 * (y / dims[1]) - 1) * h / 2  # transform y pixels to camera coords
    ray = np.array([0, -Py, -1])  # construct ray vector
    ray = ray / np.linalg.norm(ray)

    # rotate it around y based on x
    alpha = (x - (dims[0] / 2.0)) * step  # map to angle coords (-180 to 180 degrees)
    R = spatial.transform.Rotation.from_euler('Y', alpha, degrees=True).as_matrix()

    return np.dot(ray, R)  # return rotated ray

def pano_to_persp(x, y, fov, step, dims):
    """
    Transforms pixels in a panoramic image to pixel coordinates as though they had been captured using a
    perspective camera with the same dimensions.

    Args:
        x: array of pixel x-coordinates. These will be projected onto the image plane.
        y: array of pixel y-coordinate.
        fov: the camera's vertical field of view (degrees)
        step: the angular step between pixels in the x-direction.
        dims: the dimensions of the image the pixels are contained in.

    Returns:
        px, py = numpy array with the projected pixel coordinates.
    """
    # are px and py lists?
    if isinstance(x, np.ndarray) or isinstance(x, list):

        # do housekeeping to make sure arrays/lists are proper and good :)
        assert (isinstance(y, np.ndarray) or isinstance(y, list)), "Error both x and y must be lists or arrays."
        if isinstance(x, np.ndarray):
            x = x.ravel()
            y = y.ravel()

        assert len(x) == len(y), "Error, px and py must have identical dimensions."

        out = []
        for i in range(len(x)):
            out.append(pano_to_persp(x[i], y[i], fov, step, dims).ravel())
        return np.array(out)

    else:  # no - individual pixels :)

        # convert pixel to ray
        xyz = pix_to_ray_pano(x, y, fov, step, dims)

        # project ray to pixel using pinhole model
        pp, vis = proj_persp(xyz, C=np.array([0, 0, 0]),
                             a=np.array([0, 0, 0]),
                             fov=fov, dims=dims)

        return pp[:, 0:2]  # return pixel coords

def pnp(kxyz, kxy, fov, dims, ransac=True, **kwds):
    """
    Solves the pnp problem to locate a camera given keypoints that are known in world and pixel coordinates using opencv.

    Args:
        kxyz: Nx3 array of keypoint positions in world coordinates.
        kxy: Nx2 array of corresponding keypoint positions in pixel coordinates.
        fov: the (vertical) field of view of the camera.
        dims: the dimensions of the image in pixels
        ransac: true if ransac should be used to filter outliers. Default is True.
        **kwds: keyword arguments are passed to cv2.solvePnP(...) or cv2.solvePnPRansac(...).
                Other arguments can include:

            - optical_centre = the pixel coordinates (cx,cy) of the optical centre to use. By default the
                 middle pixel of the image is used. Additionally a custom optical center can be passed as (cx,cy)
    Returns:
        A tuple containing:

         - p = the camera position in world coordinates
         - r = the camera orientation (as XYZ euler angle).
         - inl = list of Ransac inlier indices used to estimate the position, or None if ransac == False.
    """
    # normalize keypoints so that origin is at mean
    mean = np.mean(kxyz, axis=0)
    kxyz = kxyz - mean

    # flip kxy coords to match opencv coord system
    # kxy[:, 1] = dims[1] - kxy[:, 1]
    # kxy[:, 0] = dims[0] - kxy[:, 0]

    # compute camera matrix
    tanfov = np.tan(np.deg2rad(fov / 2))
    aspx = dims[0] / dims[1]
    fx = dims[0] / (2 * tanfov * aspx)
    fy = dims[1] / (2 * tanfov)
    cx, cy = kwds.get("optical_centre", (0.5 * dims[0] - 0.5, 0.5 * dims[1] - 0.5))
    if 'optical_centre' in kwds: del kwds['optical_centre']  # remove keyword
    cameraMatrix = np.array([[fx, 0, cx],
                             [0, fy, cy],
                             [0, 0, 1]])

    # distortion free lens
    dist_coef = np.zeros(4)

    # use opencv to solve pnp problem and hence camera position
    if ransac:
        suc, rot, pos, inl = cv2.solvePnPRansac(objectPoints=kxyz[:, None, :3].copy(),  # points in world coords
                                                imagePoints=kxy[:, None, :2].copy(),  # points in img coords
                                                cameraMatrix=cameraMatrix,  # camera matrix
                                                distCoeffs=dist_coef,
                                                **kwds)
    else:
        inl = None
        suc, rot, pos = cv2.solvePnP(objectPoints=kxyz[:, None,  :3].copy(),  # points in world coords
                                     imagePoints=kxy[:, None, :2].copy(),  # points in img coords
                                     cameraMatrix=cameraMatrix,  # camera matrix
                                     distCoeffs=dist_coef,
                                     **kwds)

    assert suc, "Error - no solution found to pnp problem."

    # get first solution
    if not inl is None:
        inl = inl[:, 0]
    pos = np.array(pos[:, 0])
    rot = np.array(rot[:, 0])

    # apply rotation position vector
    p = -np.dot(pos, spatial.transform.Rotation.from_rotvec(rot).as_matrix())

    # convert rot from axis-angle to euler
    r = spatial.transform.Rotation.from_rotvec(rot).as_euler('xyz', degrees=True)

    # apply dodgy correction to r (some coordinate system shift)
    r = np.array([r[0] - 180, -r[1], -r[2]])

    return p + mean, r, inl



"""
Manage metadata and other header file information.
"""

import copy
import numpy as np

class HyHeader( dict ):
    """
    A generic class for encapsulating header files (.hdr) data. We use these for point clouds and spectral
    libraries as well as images.
    """
    def __init__(self):
        super().__init__()

        #default values
        self['file type'] = 'ENVI Standard'

    def has_band_names(self):
        """
        Return true if band names are defined
        """
        return 'band names' in self

    def has_wavelengths(self):
        """
        Return true if wavelengths are defined.
        """
        return ('wavelength' in self)

    def has_fwhm(self):
        """
        Return true if FWHM are defined.
        """
        return 'fwhm' in self

    def has_bbl(self):
        """
        Returns true if a bad band list has been defined.
        """
        return 'bbl' in self

    def band_count(self):
        """
        Return the number of bands specified in this header file.
        """
        if self.has_wavelengths():
            return len(self.get_wavelengths())
        elif self.has_band_names():
            return len(self.get_band_names())
        elif self.has_fwhm():
            return len(self.get_fwhm())
        elif 'bands' in self:
            return int(self['bands'])
        return 0 # image has no bands

    def get_band_names(self):
        """
        Get list of band names from header file.
        """
        assert 'band names' in self, "Error - header file has no band names."
        return self.get_list('band names', str)

    def get_wavelengths(self):
        """
        Get list of band wavelengths from header file.
        """
        if 'wavelength' in self:
            return self['wavelength'].copy()
        else:
            bc = self.band_count()
            if bc > 0:
                return np.arange(bc)
            else:
                assert False, "Error - header file has no wavelength information."


    def get_bbl(self):
        """
        Get the list of bad bands from header file. This will return an
        array scored True if a band is good and False if it is bad.
        """
        assert 'bbl' in self, "Error - header file has no bad band information."
        return np.array(self['bbl']) == 1 # internally we store as 0 = bad band and 1 = good band.

    def get_fwhm(self):
        """
        Get list of band widths from header file.
        """
        assert 'fwhm' in self, "Error - header file has no FWHM information."
        return self['fwhm'].copy()

    def set_band_names(self, band_names):
        """
        Set list of band names from header file.

        Args:
            band_names (list): a list of band names.
        """
        if band_names is None:
            if 'band names' in self:
                del self['band names']
        else:
            assert isinstance(band_names, list) or isinstance(band_names, np.ndarray), "Error - band_names must be a list or numpy array."
            self['band names'] = list(band_names)

    def set_wavelengths(self, wavelengths):
        """
        Set list of band wavelengths from header file.

        Args:
            wavelengths (list, ndarray): a list or numpy array of wavelengths.
        """
        if wavelengths is None:
            if 'wavelength' in self:
                del self['wavelength']
        else:
            assert isinstance(wavelengths, list) or isinstance(wavelengths, np.ndarray) or isinstance(wavelengths, tuple), "Error - wavelengths must be list, tuple or numpy array."
            self['wavelength'] = np.array(wavelengths).astype(float)

    def set_fwhm(self, fwhm):
        """
        Set list of band wavelengths from header file.

        Args:
            fwhm (list, ndarray): a list or numpy array of band widths.
        """
        if fwhm is None:
            if 'fwhm' in self:
                del self['fwhm']
        else:
            assert isinstance(fwhm, list) or isinstance(fwhm, np.ndarray), "Error - fwhm must be list or numpy array."
            self['fwhm'] = np.array(fwhm).astype(float)

    def set_bbl(self, bbl):
        """
        Set list of bad bands.

        Args:
            bbl (list, ndarray): a list or numpy array scored True for good bads and False for bad ones, or None to remove bbl.
        """
        if bbl is None:
            if 'bbl' in self:
                del self['bbl']
        else:
            assert isinstance(bbl, list) or isinstance(bbl, np.ndarray), "Error - bbl must be list or numpy array."
            self['bbl'] = np.array(bbl).astype(np.uint8) # internally we store as 0 (bad band) or 1 (good band).

    def drop_bands(self, mask):
        """
        Remove all header data (e.g. band names, wavelengths) associated with the specified bands. Used when removing or
        deleting bands from an image.

        Args:
            mask (list, ndarray): either a list of band indices to remove or a boolean numpy array with True for bands that should remove.
        """
        # how many bands are there?
        nbands = self.band_count()

        # create boolean mask
        mask = np.array(mask)
        if mask.dtype == bool:
            keep = np.logical_not(mask)  # flip so we have array with True for bands to keep
        elif issubclass(mask.dtype, np.integer):
            keep = np.full(nbands, True)
            keep[mask] = False  # drop bands
        else:
            "Error - unrecognised data type %s." % mask.dtype

        # check we have a value per band
        #assert len(keep) == nbands, "Error - data has %d values but mask is length %d" % (nbands, len(keep))
        # N.B. this assert has been removed because it is triggered when a header contains no wavelength or band names
        # (and so the header file cannot know how many bands to expect in the image)

        skip = ['default bands', 'temperature', 'camera',
                'class lookup', 'class names', 'coordinate system string', 'description',
                'geo points', 'map info', 'projection info', 'spectra names']
        for key, value in self.items():
            if key in skip:  # these keys are lists but should never be trimmed (as they don't relate to bands)
                continue
            if isinstance(value, str):
                if "," in value:
                    value = np.array(value.split(","))  # split on commas
            if isinstance(value, list):
                if len(value) == len(keep):
                    self[key] = list(np.array(value)[keep])  # drop bands
            if isinstance(value, np.ndarray):
                if len(value) == len(keep):
                    self[key] = value[keep]  # drop bands

        #update band count
        self['bands'] = str(np.sum(keep))

    def drop_all_bands(self):
        """
        Remove all attributes in the header file that are specified per band. Useful if a dataset is copied and the bands
        overwritten.
        """

        # how many bands are there?
        nbands = self.band_count()
        if nbands == 0:
            return # no bands

        skip = ['default bands', 'temperature', 'camera',
                'class lookup', 'class names', 'coordinate system string', 'description',
                'geo points', 'map info', 'projection info', 'spectra names']
        to_del = []
        for key, value in self.items():
            if key in skip:  # these keys are lists but should never be trimmed (as they don't relate to bands)
                continue
            if isinstance(value, str):
                if "," in value:
                    value = np.array(value.split(","))  # split on commas
            if isinstance(value, list):
                if len(value) == nbands:
                    to_del.append(key)
            if isinstance(value, np.ndarray):
                if len(value) == nbands:
                    to_del.append(key)

        # do delete
        for k in to_del:
            del self[k]

        #update band count
        self['bands'] = str(0)

    def copy(self):
        """
        Make a deep copy of this header class.

        Returns:
            a new header instance.
        """

        return copy.deepcopy(self)

    def print(self):
        """
        Print metadata to console
        """

        for key, value in self.items():
            s = str(value).strip()
            print("%s = { %s } " % (key, s))

    ############################################################
    ## Functions related to other data we store in header files
    ############################################################
    def get_list(self, name, dtype=float):
        """
        Get generic list data from the header as a numpy array.

        Args:
            name (str): the header key used to access the list.
            dtype (str): the numpy data type that should be returned. Default is float.

        Returns:
            a numpy array containing the header data, or None if the key does not exist.
        """

        if not name in self:
            return None
        out = self[name]
        if isinstance(out, str):
            out = np.array( [v.strip() for v in out.split(',')] ) # split string into array
        try:
            return out.astype(dtype) # convert to the correct type and return
        except:
            return out # return as e.g. string - is better than failing.

    def set_camera(self, camera, n=0):
        """
        Store information on the position and orientation of this image to the image header.

        Args:
            camera (hylite.project.Camera): the camera instance to store.
            n (int): the camera id (integer), if multiple cameras are stored (e.g. for hyperclouds). Default is 0.
        """

        # pose = cx cy cz rx ry rz projection
        self["camera %d pose" % n] = "%f %f %f %f %f %f" % (camera.pos[0], camera.pos[1], camera.pos[2],
                                                     camera.ori[0], camera.ori[1], camera.ori[2])

        # internals = x pixels, y pixels, vertical field of view, angular step (for panoramic cameras)
        if camera.is_panoramic():
            self["camera %d internals" % n] = "%s %d %d %f %f" % (
            camera.proj, camera.dims[0], camera.dims[1], camera.fov, camera.step)
        else:
            self["camera %d internals" % n] = "%s %d %d %f" % (camera.proj, camera.dims[0], camera.dims[1], camera.fov)

    def get_camera(self, n = 0):
        """
        Extract a camera object containing the internal and external camera properties (or None if they don't exist).

        Args:
            n (int): the camera id (integer), if multiple cameras are stored (e.g. for hyperclouds). Default is 0.
        """
        from hylite.project.camera import Camera #n.b. do import here to avoid circular dependencies

        # check for single camera (for backward compatability)
        if 'camera pose' in self and 'camera internals' in self:
            self['camera %d pose'%n] = self['camera pose']
            self['camera %d internals'%n] = self['camera internals']
            del self['camera pose']
            del self['camera internals']

        # no camera info? bail!
        if not ('camera %d pose'%n in self and 'camera %d internals'%n in self):
            return None

        pose = self['camera %d pose' % n].split(' ')
        internal = self["camera %d internals" % n].split(' ')
        pos = np.array([float(pose[0]), float(pose[1]), float(pose[2])])
        ori = np.array([float(pose[3]), float(pose[4]), float(pose[5])])
        proj = internal[0]
        dims = (int(internal[1]), int(internal[2]))
        fov = float(internal[3])
        if 'pano' in proj:
            step = float(internal[4])
            return Camera(pos=pos, ori=ori, proj=proj, fov=fov, dims=dims, step=step)
        elif 'persp' in proj:
            return Camera(pos=pos, ori=ori, proj=proj, fov=fov, dims=dims)
        else:
            assert False, "Error - '%s' is an unknown camera projection." % proj

    def has_camera(self):
        """
        Return true if this header file has defined camera pose information.
        """
        return not (self.get_camera() is None)

    def get_panel_names(self):
        """
        Get the name of all panels defined in this header file.
        """
        names = []
        for k in self.keys():
            if "target" in k: # this key relates to a target
                name = k.split(' ')[1]
                if not name in names:
                    names.append(name)
        return names

    def add_panel(self, panel, name=None):
        """
        Add the specified panel to  the header file

        Args:
            panel (hylite.correct.Panel): the panel reflectance and measured radiance spectra.
            name (str): the name of this panel. If None (default) then the name of the panel material will be used (e.g., 'R90').
        """
        assert (panel.get_wavelengths() == self.get_wavelengths()).all(), "Error - target wavelengths and header wavelengths do not match."
        if name is None:
            name = panel.material.get_name()
        name = name.lower()  # use lower case
        self['target %s reflectance' % name] = panel.get_reflectance()
        self['target %s radiance' % name] = panel.get_mean_radiance()
        if panel.normal is not None: # store normal vector
            self['target %s normal' % name] = panel.normal
        if panel.skyview is not None:
            self['target %s skyview' % name] = panel.skyview
        if panel.alpha is not None:
            self['target %s alpha' % name] = panel.alpha

    def get_panel(self, name):
        """
        Get the specified panel from this header

        Args:
            name (str): the name of the target to return.

        Returns:
             a Panel instance containing the reflectance and radiance data, or None if it does not exist.
        """

        from hylite.reference.spectra import Target # n.b. do import here to avoid circular dependencies
        from hylite.correct.panel import Panel

        name = name.lower() # use lower case
        if not ('target %s reflectance' % name in self and 'target %s radiance' % name in self):
            return None # no target found

        reflectance = self['target %s reflectance' % name]
        radiance = self['target %s radiance' % name]

        if isinstance(reflectance,str):
            reflectance = np.fromstring(reflectance, sep=",")
            radiance = np.fromstring(radiance, sep=",")

        reflectance = reflectance.astype(np.float32) # check type
        radiance = radiance.astype(np.float32)

        # create dummy Target instance
        material = Target(self.get_wavelengths(), reflectance, name=name)

        # create Panel instance and return
        P = Panel( material, radiance, wavelengths=self.get_wavelengths() )

        # get normal vector (if defined)
        if ('target %s normal' % name) in self:
            normal = self['target %s normal' % name]
            if isinstance(normal, str):  # parse string
                P.set_normal( np.fromstring(self['target %s normal' % name], sep=",") )
            elif isinstance(normal, list):
                P.set_normal( np.array(normal) )
            else:
                assert isinstance(normal, np.ndarray), "Error - %s is an invalid normal." % normal

        # get illu factors if defined
        if ('target %s alpha' % name) in self:
            P.alpha = float(self['target %s alpha' % name])
        if ('target %s skyview' % name) in self:
            P.skyview = float(self['target %s skyview' % name])
        return P

    def remove_panel(self, name = None):
        """
        Removes the specified calibration panel from this header.

        Args:
            name (str): the name of the panel to remove. If None (default) then all panels will be removed.
        """
        if name is not None:
            name = name.lower()  # use lower case

        # find keys to remove
        to_del = []
        for k in self.keys():
            if "target" in k: # this key relates to a target
                if not name is None and not name in k:
                    continue # skip this one
                to_del.append(k)

        # remove them
        for k in to_del:
            self.pop(k, None)

    def get_data_ignore_value(self ):
        if 'data ignore value' in self:
            return float(self['data ignore value'])
        else:
            return np.nan

    def set_data_ignore_value(self, val):
        assert isinstance(val, float) or isinstance(val, int), "Error - %s is an invalid data type for data ignore values" % type(val)
        self['data ignore value'] = str(val)

    def get_class_names(self):
        if not 'class names' in self:
            return None
        if isinstance(self['class names'], list):
            return self['class names']
        elif isinstance(self['class names'],str):
            return [s.strip() for s in self['class names'].split(',')]

    def get_sample_points(self, name):
        """
        Get points that are associated with the specified sample name.
        """
        key = 'sample %s' % name
        if not key in self:
            key = 'sample %s' % name.lower()
            if not key in self:
                return []

        if isinstance(self[key], list):
            return self[key]
        elif isinstance(self[key], str):
            if ')' in self[key]: # image coordinates - parse tuple
                # parse string
                pnts = []
                splt = self[key].split('),')
                for tpl in splt:
                    tpl = tpl.replace('(', '')
                    tpl = tpl.replace(')', '')
                    tpl = tpl.split(',')
                    pnts.append((int(tpl[0]), int(tpl[1])))
            else:
                splt = self[key].split(',') # point cloud - simply treat as IDs
                pnts = (int(i) for i in splt)
            return pnts

    def set_sample_points(self, name, points):
        """
        Set points that are associated with the specified sample name.

        Args:
            name (str): the name of the sample points to set
            points (list): a list of points: [(x1,y1), (x2,y2) ... ] or [id1,id2,id3, ...]
        """
        key = 'sample %s' % name
        self[key] = points



"""
Hyperspectral detrending algorithms such as hull correction.
"""

from hylite import HyLibrary, HyData
import numpy as np
from gfit.util import remove_hull

def polynomial(data, degree = 1, method='div'):

    """
    Detrend an image data array using a polynomial fit and np.polyfit( ... ).

    Args:
        data: numpy array of the format image[x][y][b].
        degree: the degree of the polynomial to fit. Default is 2.
        method: 'divide' or 'subtract'. Default is 'divide'.

    Returns:
        A tuple containing:

         - corr = the corrected (detrended) data.
         - trend = the trend that was removed.
    """

    #calculate trend
    y = np.array(data.reshape( -1, data.shape[-1] )) #reshape array so each pixel is a column
    y[np.logical_not(np.isfinite(y))] = 0 #kill NaNs
    _x = np.arange(data.shape[-1])
    fit = np.polynomial.polynomial.polyfit(_x, y.T, degree) # fit polynomial
    t = np.polynomial.polynomial.polyval(_x, fit) # evaluate it

    #apply correction
    if 'div' in method:
        y /= t
    elif 'sub' in method.lower():
        y -= t
        y += np.min(y) #map to positive

    return y.reshape(data.shape), t.reshape(data.shape)


def get_hull_corrected(data, band_range=None, method='div', hull='upper', vb=True):
    """
    Apply a hull correction to an entire HyData instance (HyImage, HyCloud or HyLibrary). Returns a corrected copy of
    the input dataset. Note that noise can greatly effect hull corrections, so you should consider denoising first (see
    HyData.smooth_median(...) and HyData.smooth_savgol(...).

    Args:
        data: a numpy array or HyData instance to detrend.
        band_range: Tuple containing the (min,max) band indices or wavelengths to run the correction between. If None
                     (default) then the correction is run of the entire range. Only works if data is a HyData instance.
        method: Trend removal method: 'divide' or 'subtract'. Default is 'divide'.
        hull: 'upper' if a hull should be fitted to the top of the data (default), or 'lower' if it should be fit to the bottom of the data.
        vb: True if this should print output.

    Returns:
        A hull corrected dataset.
    """

    if isinstance(data, HyData):

        # create copy containing the bands of interest
        if band_range is None:
            band_range = (0, -1)
        else:
            band_range = (data.get_band_index(band_range[0]), data.get_band_index(band_range[1]))
        corrected = data.export_bands(band_range)

        # convert integer data to floating point (we need floats for the hull correction)
        comp = False
        if corrected.is_int():
            if np.nanmax( corrected.data ) > 100: # check large number used in compressed form
                corrected.decompress()
            else:
                corrected.data = corrected.data.astype(np.float32) # cast to float for hull correction
            comp = True

        # get valid pixels
        D = corrected.get_raveled()
        nan = corrected.header.get_data_ignore_value()
    else:
        shape = data.shape
        D = data.reshape((-1, shape[-1]))
        nan = np.inf

    # check values are all positive - negative values break the hull correction!
    D = np.clip(D, 0, np.inf )

    #valid = (np.isfinite(D) & (D != nan)).all(axis=1)  # drop nans/no-data values
    D = np.nan_to_num(D, nan=0, posinf=0, neginf=0) # ensure all values are finite (and do not effect hul)
    valid = (D != D[:, 0][:, None]).any(axis=1)  # drop flat spectra (e.g. all zeros)
    if len(valid) == 0:
        return corrected  # quick exit for empty images

    # do hull correction
    if 'upper' in hull.lower():
        X = remove_hull( D[valid],upper=True, div=('div' in method), vb=vb)
    elif 'lower' in hull.lower():
        X = remove_hull(np.max(D[valid]) - D[valid], upper=True, div=('div' in method), vb=vb)
    else:
        assert False, "Error = 'hull' should be 'upper' or 'lower', not %s." % hull
    D[valid] = X  # copy data back into original array

    # do housekeeping and return
    if isinstance(data, HyData):
        if isinstance(corrected, HyLibrary):
            corrected.upper = None # reset upper and lower limits as these will now be incorrect
            corrected.lower = None # reset upper and lower limits as these will now be incorrect
        corrected.set_raveled(D)
        # convert back to integer if need be
        if comp:
            corrected.compress()
        return corrected
    else:
        return D.reshape( shape )


"""
This package contains reference features such as typical mineral absorbtions or absorbtion positions of QAQC targets.
"""
from hylite.hyfeature import *

##############################
##Create basic/common features
##############################
#N.B. All values here are approximate and largely for (1) plotting and (2) as initial values for feature fitting
#     routines. Believe nothing!
class Features:
    """
    Specific absorption types. Useful for plotting etc. Not really used for anything and will probably be depreciated soon.
    """

    H2O = [ HyFeature("H2O", p, w, color='skyblue') for p,w in [(825,50), (940,75), (1130,100), (1400,150), (1900,200), (2700,150)] ]
    OH = [HyFeature("OH", 1400, 50, color='aquamarine'), HyFeature("OH", 1550, 50, color='aquamarine'), HyFeature("OH", 1800, 100, color='aquamarine')]
    AlOH = [HyFeature("AlOH", 2190, 60, color='salmon')]
    FeOH = [HyFeature("FeOH", 2265, 70, color='orange')]
    MgOH = [HyFeature("MgOH", 2330, 60, color='blue'), HyFeature("MgOH", 2385, 30,color='blue')]
    MgCO3 = [HyFeature("MgCO3", 2320, 20, color='green')]
    CaCO3 = [HyFeature("CaCO3", 2340, 20, color='blue')]
    FeCO3 = [HyFeature("FeCO3", 2350, 20, color='steelblue')]
    Ferrous = [HyFeature("Fe2+", 1000, 400, color='green')]
    Ferric = [HyFeature("Fe3+", 650, 170, color='green')]

    # REE features
    Pr = [HyFeature("Pr", w, 5, color=(74 / 256., 155 / 256., 122 / 256., 1)) for w in [457, 485, 473, 595, 1017] ]
    Nd = [HyFeature("Nd", w, 5, color=(116 / 256., 114 / 256., 174 / 256., 1)) for w in [430, 463, 475, 514, 525, 580, 627, 680, 750, 800, 880, 1430, 1720, 2335, 2470]]
    Sm = [HyFeature("Sm", w, 5, color=(116 / 256., 114 / 256., 174 / 256., 1)) for w in [945, 959, 1085, 1235, 1257, 1400, 1550]]
    Eu = [HyFeature("Eu", w, 5, color=(213 / 256., 64 / 256., 136 / 256., 1)) for w in [385, 405, 470, 530, 1900, 2025, 2170, 2400, 2610]]
    Dy = [HyFeature("Dy", w, 5, color=(117 / 256., 163 / 256., 58 / 256., 1)) for w in [368, 390, 403, 430, 452, 461, 475, 760, 810, 830, 915, 1117, 1276, 1725]]
    Ho = [HyFeature("Ho", w, 5, color=(222 / 256., 172 / 256., 59 / 256., 1)) for w in [363, 420, 458, 545, 650, 900, 1130, 1180, 1870, 1930, 2005]]
    Er = [HyFeature("Er", w, 5, color=(159 / 256., 119 / 256., 49 / 256., 1)) for w in [390, 405, 455, 490, 522, 540, 652, 805, 985, 1485, 1545]]
    Tm = [HyFeature("Tm", w, 5, color=(102 / 256., 102 / 256., 102 / 256., 1)) for w in [390, 470, 660, 685, 780, 1190, 1640, 1750]]
    Yb = [HyFeature("Yb", w, 5, color=(209 / 256., 53 / 256., 43 / 256., 1)) for w in [955, 975, 1004 ]]

# common minerals
class Minerals:
    """
    Common mineral absorption features. Useful for plotting etc. Not really used for anything and will probably be depreciated soon.
    """

    # Kaolin clays (dominant SWIR feature)
    Kaolinite = [HyFeature("Kaolinite/Halloysite", 2200, 100, color='aquamarine')]
    Halloysite = [Kaolinite]
    Dickite = [HyFeature("Dickite/Nacrite", 2180, 100, color='aquamarine')]
    Nacrite = [Dickite]
    KAOLIN = MultiFeature("Kaolin", Kaolinite + Dickite)

    Pyrophyllite = HyFeature("Pyrophyllite", 2160.0, 150, color='aquamarine')

    #Smectite clays (dominant SWIR feature)
    Montmorillonite = [HyFeature("Montmorillonite", 2210.0, 125, color='orange')]
    Nontronite = [HyFeature("Nontronite", 2280, 125, color='orange')]
    Saponite = [HyFeature("Saponite", 2309, 100, color='orange')]
    SMECTITE = MultiFeature("Smectite", Montmorillonite + Nontronite + Saponite)

    # white micas (dominant SWIR feature)
    Mica_Na = [HyFeature("Mica (Na)", 2150, 150, color='coral' )]
    Mica_K = [HyFeature("Mica (K)", 2190, 150, color='lightcoral' )]
    Mica_MgFe = [HyFeature("Mica (Mg, Fe)", 2225, 150 , color='sandybrown')]
    MICA = MultiFeature("White mica", Mica_Na + Mica_K +Mica_MgFe)

    # chlorite
    Chlorite_Mg = [ HyFeature("Chlorite (Mg)", 2245.0, 50, color='seagreen'), HyFeature("Chlorite (Mg)", 2325.0, 50, color='seagreen') ]
    Chlorite_Fe = [HyFeature("Chlorite (Fe)", 2261.0, 50, color='seagreen'), HyFeature("Chlorite (Fe)", 2355.0, 50, color='seagreen') ]
    CHLORITE = [MultiFeature("Chlorite (FeOH)", [Chlorite_Mg[0], Chlorite_Fe[0]]),
                MultiFeature("Chlorite (MgOH)", [Chlorite_Mg[1], Chlorite_Fe[1]])]

    # biotite
    Biotite_Mg = [ HyFeature("Biotite (Mg)", 2326, 50, color='firebrick'), HyFeature("Biotite (Mg)", 2377, 50, color='firebrick') ]
    Biotite_Fe = [ HyFeature("Biotite (Fe)", 2250, 50, color='firebrick'), HyFeature("Biotite (Fe)", 2350, 50, color='firebrick') ]
    BIOTITE = [MultiFeature("Biotite (FeOH)", [Biotite_Mg[0], Biotite_Fe[0]]),
               MultiFeature("Biotite (MgOH)", [Biotite_Mg[1], Biotite_Fe[1]]) ]

    # amphiboles Tremolite, hornblende, actinolite
    Amphibole_Mg =  [HyFeature("Amphibole (Mg)", 2320.0, 50, color='royalblue')]
    Amphibole_Fe =  [HyFeature("Amphibole (Fe)", 2345.0, 50, color='royalblue')]
    AMPHIBOLE = MultiFeature("Amphibole", Amphibole_Mg + Amphibole_Fe)

    # carbonate minerals
    Dolomite = [HyFeature("Dolomite", 2320, 20, color='green')]
    Calcite = [HyFeature("Calcite", 2345, 20, color='blue')]
    Ankerite = [HyFeature("Ankerite", 2330, 20, color='steelblue')]
    CARBONATE = MultiFeature("Carbonate", Dolomite+ Ankerite+ Calcite)

    #Sulphates Jarosite
    Gypsum = [HyFeature("Gypsum", 1449.0, 50, color='gold'), HyFeature("Gypsum", 1750, 50, color='gold'), HyFeature("Gypsum", 1948.0, 50, color='gold')]
    Jarosite = [HyFeature("Jarosite", 1470.0, 50, color='orange'), HyFeature("Jarosite", 1850, 50, color='orange'), HyFeature("Jarosite", 2270.0, 50, color='orange')]
    SULPHATE = MultiFeature( "Sulphate", Gypsum + Jarosite )

    # misc
    Epidote = [ HyFeature("Epidote", 2256.0, 40, color='green'), HyFeature("Epidote", 2340.0, 40, color='green')]

# and some useful 'themes' (for plotting etc)
class Themes:
    """
    Some useful 'themes' (for plotting etc)
    """
    ATMOSPHERE = Features.H2O  #[HyFeature("H2O", 975, 30), HyFeature("H2O", 1395, 120), HyFeature("H2O", 1885, 180), HyFeature("H2O", 2450, 100)]
    CARBONATE = [Minerals.CARBONATE]
    OH = Features.AlOH + Features.FeOH + Features.MgOH
    CLAY = [ Minerals.KAOLIN, Minerals.SMECTITE ]
    DIAGNOSTIC = Features.Ferrous + Features.AlOH+Features.FeOH+Features.MgOH

#expose through HyFeature class for convenience
HyFeature.Features = Features
HyFeature.Minerals = Minerals
HyFeature.Themes = Themes



"""
Classes containing different occlusion models (e.g. estimators for shadows, sky view factor etc.)
"""

import numpy as np
from hylite.analyse import band_ratio


def calcBandRatioOcc(rad, **kwds):
    """
    Calculate the fraction of incoming light blocked by this occlusion (0 - 1) based on a band ratio.

    Args:
        radiance: Radiance data (HyData instance) to calculate band ratio with.
        **kwds: Keywords can include:

             - num = Numerator of band ratio, as defined by hylite.analyse.band_ratio. Default is (400,450) nm.
             - den = Denominator of band ratio, as defined by hylite.analyse.band_ratio. Default is (550,600) nm.
             - a = the threshold for complete shadow (values > this are full shadow). Default is 1.20.
             - b = the threshold for complete sun (values < this are full sun). Default is 1.3.
    Returns:
        A HyData instance containing the occlusion factor (0-1) for each point.
    """
    br = band_ratio(rad, kwds.get('num', (400., 450.)), kwds.get('num', (550., 600.)))
    out = br.data[..., 0] - kwds.get('a', 1.20)
    out /= kwds.get('b', 1.3) - kwds.get('a', 1.2)
    out = np.clip(out, 0, 1)
    return out

def calcProjectedOcc(self, geom, sunvec, s=3 ):
    """
    Calculate projected shadows based on the scene geometry and sun direction vector.

    Args:
        geom: a HyCloud instance describing scene geometry.
        sunvec: a (3,) numpy array containing the downward pointing illumination direction.
        s: the size of points in the cloud.
    """
    assert False, "Error: this function has not been implemented yet."




"""
Classes containing different occlusion models (e.g. estimators for shadows, sky view factor etc.)
"""
import hylite
import matplotlib.pyplot as plt
import numpy as np

from hylite.correct import get_hull_corrected
from hylite.multiprocessing import parallel_chunks


def estimate_path_radiance(image, depth, thresh=1):
    """
    Apply the dark object subtraction (DOS) method to estimate path radiance in the provided image.

    Args:
        image: the hyperspectral image (HyImage instance) to estimate path radiance for.
        depth: A 2-D (width,height) numpy array of pixel depths in meters. This can be easily computed using
               a HyScene instance.
        thresh: the percentile threshold to use when selecting dark pixels. Default is 1%.

    Returns:
        A tuple containing:

            - spectra = a numpy array containing the estimated path radiance spectra (in radiance per meter of depth).
            - path = a HyImage instance containing the estimated path radiance per pixel (computed by multiplying
                  the spectra by the depth).
    """

    # identify dark pixels
    r = np.nanmean(image.data, axis=-1)  # calculate mean brightness

    assert r.shape == depth.shape, "Error: depth array and HSI have different shapes: %s != %s" % (
    str(r.shape), str(depth.shape))

    r[r == 0] = np.nan  # remove background / true zeros as this can bias the percentile clip
    r[np.logical_not(np.isfinite(depth))] = np.nan  # remove areas without depth info
    r[depth == 0] = np.nan  # remove areas without depth info

    # extract dark pixels and get median
    thresh = np.nanpercentile(r, thresh)  # threshold for darkest pixels
    darkref = image.data[r <= thresh, :]
    darkdepth = depth[r <= thresh]  # corresponding depths

    # compute path radiance estimate
    S = darkref / darkdepth[..., None]  # compute estimated path-radiance per meter
    S = np.nanpercentile(S, 50, axis=0)  # take median of all estimates

    # compute per pixel path radiance
    P = image.copy(data=False)
    P.data = depth[..., None] * S[None, None, :]

    return S, P

def correct_path_absorption(data, band_range=(0, -1), thresh=99, atabs = 1126., vb=True):
    """
    Fit and remove a known atmospheric water feature to remove atmospheric path absorbtions from reflectance spectra.
    See  Lorenz et al., 2018 for more details.

    Reference:
    https://doi.org/10.3390/rs10020176

    Args:
        image: a hyperspectral image to correct
        band_range: a range of bands to do this over. Default is (0,-1), which applies the correction to all bands.
        thresh: the percentile to apply when identifying the smallest absorbtion in any range based on hull corrected
                spectra. Lower values will remove more absorption (potentially including features of interest).
        atabs: wavelength position at which a known control feature is situated that defines the intensity of correction
                - for atmospheric effects, this is set to default to 1126 nm
        vb: True if a progress bar should be created during hull correction steps.

    Returns:
        a HyData instance containing the corrected spectra.
    """
    assert isinstance(atabs, float), "Absorption wavelength must be float"
    # subset dataset
    out = data.export_bands(band_range)
    nanmask = np.logical_not(np.isfinite(out.data))
    out.data[nanmask] = 0  # replace nans with 0

    # get depth of water feature at 1126 nm for all pixels
    atm_depth = (out.get_band(atabs - 50.) + out.get_band(atabs + 50.)) / 2 - out.get_band(atabs)

    # kick out data points with over-/undersaturated spectra
    atm_temp = atm_depth.copy()
    atm_temp[np.logical_or(np.nanmax(out.data, axis=-1) >= 1, np.nanmax(out.data, axis=-1) <= 0)] = 0
    # extract pixels that are affected most by the features
    highratio = out.data[atm_temp > np.percentile(atm_temp, 90)]
    # hull correct those
    hull = get_hull_corrected(hylite.HyData(highratio), vb=vb)
    # extract the always consistent absorptions
    hull_max = np.nanpercentile(hull.data, thresh, axis=0)

    vmin = hull_max[out.get_band_index(atabs)]
    # apply adjustment and return
    if out.is_image():
        nmin = -atm_depth[..., None]
        out.data -= ((hull_max[None, None, :] - vmin) * (-nmin) / (1 - vmin) + nmin)
    else:
        nmin = -atm_depth[..., None]
        out.data -= ((hull_max[None, :] - vmin) * (-nmin) / (1 - vmin) + nmin)
    out.data[nanmask] = np.nan  # add nans back in
    out.data = np.clip(out.data, 0, 1)

    return out


"""
This package implements common analysis methods for hyperspectral datasets, including:

- supervised and unsupervised classification.
- minimum wavelength mapping.
- endmember extraction and unmixing.
- commonly used band ratios (e.g. NDVI).
"""

from .indices import *
from .mwl import *
from .sam import *
from .dtree import *

import matplotlib.pyplot as plt

def saveLegend(red: str, green: str, blue: str, path: str):
    """
    Create and save a legend image for a ternary image.

    Args:
        - red: String containing the description / label for the red channel.
        - green: String containing the description / label for the green channel.
        - blue: String containing the description / label for the blue channel.
        - path: The output path to save the image to.
    """

    tr = [np.nan, np.nan, np.nan]
    plt.figure(figsize=(4, 0.5))
    plt.fill(tr, tr, label=red, color='r')
    plt.fill(tr, tr, label=green, color='g')
    plt.fill(tr, tr, label=blue, color='b')
    plt.legend(ncol=3, loc='center')
    plt.axis('off')
    os.makedirs(os.path.dirname(path), exist_ok=True)
    plt.tight_layout()
    plt.savefig(path, dpi=350)
    plt.close()



import unittest, os
from hylite import io
from pathlib import Path
import numpy as np

class MyTestCase(unittest.TestCase):
    def test_correct_path_absorption(self):
        from hylite.correct.illumination import correct_path_absorption
        image = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))
        cloud = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"), "hypercloud.hdr") )
        # test hull correction on HyData instances
        for D in [image, cloud]:
            Xhc = correct_path_absorption(D, atabs=2200.,vb=False)
            self.assertEqual(D.data.shape, Xhc.data.shape)
            self.assertGreaterEqual(np.nanmin(Xhc.data), 0.0)
            self.assertLessEqual(np.nanmax(Xhc.data), 1.0)



if __name__ == '__main__':
    unittest.main()



"""
Utility functions for combining / averaging multiple Datasets together.
"""

import numbers
import cv2
import numpy as np
from hylite import io as io


def combine(image_list, method="median", warp=False):
    """
    Combines a list of images to improve the signal to noise ratio/remove issues associated with dead pixels etc.

    Args:
        image_list: a list of hyImage objects for which data will be averaged. These must be identical sizes.
        method: The method used to combine the images. Can be "mean", "median", "min", "max" or a
                percentile between 0 and 100. Default is "median"
        warp: should the images be warped to optimise coregistration using optical flow? Slow... default is False. Always
              matches images to the first one in image_list. The middle band is used for matching.
    Returns:
        A tuple containing:

         - average = a numpy array containing the averaged image data
         - std = a numpy array containing the standard deviation of the image data.
    """

    # resize to fit (sometimes some images are 1-2 pixels too long for line-scanners on a tripod)
    minx = min([i.xdim() for i in image_list])
    miny = min([i.ydim() for i in image_list])
    arr = [i.data[0:minx, 0:miny, :] for i in image_list]

    # build co-aligned data array to average
    if warp:
        alg = cv2.optflow.createOptFlow_DeepFlow()
        X, Y = np.meshgrid(range(arr[0].shape[1]), range(arr[0].shape[0]))
        match_idx = int(image_list[0].band_count() / 2)
        bnd1 = io.HyImage.to_grey(arr[0][:, :, match_idx])  # we use the middle band to do alignment

        for i, image in enumerate(arr[1:]):
            print("Warping image %d" % (i + 1))

            # extract matching band
            bnd2 = io.HyImage.to_grey(image[:, :, match_idx])

            # calculate optical flow
            flow = alg.calc(bnd1, bnd2, None)

            # transform to pixel map
            map = np.dstack([X, Y]).astype(np.float32)
            map[:, :, 0] += flow[:, :, 0]
            map[:, :, 1] += flow[:, :, 1]

            # remap bands
            for b in range(image.shape[-1]):
                image[:, :, b] = cv2.remap(image[:, :, b], map, None, cv2.INTER_LINEAR)

    # calculate average and standard deviation
    std = np.nanstd(arr, axis=0)
    if "mean" in method.lower():
        out = np.nanmean(arr, axis=0)
    elif "median" in method.lower():
        out = np.nanmedian(arr, axis=0)
    elif "min" in method.lower():
        out = np.nanmin(arr, axis=0)
    elif "max" in method.lower():
        out = np.nanmax(arr, axis=0)
    elif isinstance(method, numbers.Number) and 0 <= method <= 100:  # percentile
        out = np.nanpercentile(arr, axis=0)

    # return result
    return out, std


import unittest
import os
from pathlib import Path
import hylite.io as io
from hylite.project.align import *
import shutil
from tempfile import mkdtemp
import numpy as np

class TestAlign(unittest.TestCase):
    def test_library(self):
        image = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))

        # label some seed pixels in each sample
        image.header.set_sample_points('A', [(20, 15)])
        image.header.set_sample_points('B', [(80, 15)])
        image.header.set_sample_points('C', [(140, 15)])
        image.header['class names'] = ['A', 'B', 'C']

        # test building library with sample positions only
        from hylite.hylibrary import from_indices
        lib = from_indices(image,
                             [image.header.get_sample_points(n)[0] for n in image.header.get_class_names()],
                             names=image.header.get_class_names(),
                             s=5)
        self.assertEqual( lib.data.shape[0], 3 )

        # expand labels using grab-cut
        from hylite.filter import label_blocks
        cls = label_blocks(image, s=5,  # number of pixels to label outside of seed point
                           epad=1,  # ignore these pixels near edges (can be dodgy sometimes)
                           erode=1,  # apply erode filter to avoid pixels near sample edges
                           boost=10,  # boost contrast before labelling
                           bands=hylite.SWIR)


        # test building library and plotting functions
        from hylite.hylibrary import from_classification
        for sample in [50, (50,),'all', (5,50,95)]:
            lib = from_classification( image, cls, ignore=[0], subsample=sample )
            lib.quick_plot(color=['r','g','b'], clip=(0,50,100))
            lib.quick_plot(color=['r', 'g', 'b'], clip=50)

        # test copy functions (well... run them)
        lib2 = lib.copy(data=False)
        lib2 = lib.copy(data=True)

        # test merging / splitting
        lib.set_sample_names(['A','B','C'])
        lib2 = lib[['A','A','B']] # check merging of names
        self.assertEqual( lib2.data.shape[0], 2 )
        self.assertEqual(lib2.data.shape[1], 6)
        lib2 = lib[[1, 'A', 'B']] # check indices and names merge seamlessly
        self.assertEqual(lib2.data.shape[0], 2)
        self.assertEqual(lib2.data.shape[1], 6)

        # add groups
        lib.add_group("Group1", ['A','B'])
        lib.add_group("Group2", ['C'])

        # get them
        groups = lib.get_groups()
        lib2 = lib.get_group('Group1')
        lib3 = lib.get_group('Group2')
        self.assertTrue( 'Group1' in groups )
        self.assertTrue('Group2' in groups )
        self.assertTrue( lib.has_groups() )
        self.assertEqual( lib2.data.shape[0], 2)
        self.assertEqual(lib3.data.shape[0], 1)

        # test collapse
        self.assertEqual(lib.data.shape[0],3) # three samples
        lib3 = lib.collapse()
        self.assertEqual(lib3.data.shape[0],2) # two samples after collapsing groups

        # test squash (at least, run it)
        lib4 = lib.squash()
        self.assertEqual( lib4.data.shape[1], 1 )
        self.assertEqual( lib4.data.shape[0], lib.data.shape[0] )

        # test fancy plotting
        lib.quick_plot( collapse=True, hc=True )

    def test_construction(self):
        image = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"), "image.hdr"))




if __name__ == '__main__':
    unittest.main()



"""
Store projection maps binding image data to point cloud data during e.g. the correction of pushbroom UAV datasets.
"""

import numpy as np
import os

from hylite.project import PMap

def savePMap(path, pmap):
    """
    Save a PMap instance using numpy.

    Args:
        path: the path to save to.
        pmap: the PMap instance to save.
    """
    pnt,pix,z = pmap.get_flat()
    dims = np.array([pmap.xdim, pmap.ydim, pmap.npoints])
    np.savez_compressed( path, dims=dims, points=pnt, pixels=pix, depths=z )

def loadPMap(path):
    """
    Load a PMap instance using numpy.

    Args:
        path: the file path to load from.

    Returns:
        a PMap instance loaded from the file.
    """

    # check extension
    if not os.path.exists(path):
        path += ".npz" # try adding npz extension to see if that helps

    assert os.path.exists(path), "Error - file not found: %s" % path

    # load data
    data = np.load( path )

    # check attributes
    if 'dims' not in data:
        data.close()
        assert False, "Error - npz does not contain a 'dims' attribute."
    if 'points' not in data:
        data.close()
        assert False, "Error - npz does not contain a 'points' attribute."
    if 'pixels' not in data:
        data.close()
        assert False, "Error - npz does not contain a 'pixels' attribute."
    if 'depths' not in data:
        data.close()
        assert False, "Error - npz does not contain a 'depths' attribute."

    # extract attrubites
    xdim, ydim, npoints = data[ "dims" ]
    points = data[ "points" ]
    pixels = data[ "pixels" ]
    depths = data[ "depths" ]

    # close file
    data.close()

    # create new PMap and populate with data
    pm = PMap( xdim, ydim, npoints )
    pm.set_flat( points, pixels, depths )
    return pm


import unittest
import os
from hylite import io
from pathlib import Path
from hylite.analyse.mwl import *
import shutil
from tempfile import mkdtemp
import numpy as np


class MyTestCase(unittest.TestCase):
    def test_hull(self):
        from hylite.correct import get_hull_corrected
        image = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))
        cloud = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))

        # test hull correction on numpy array
        Xhc = get_hull_corrected( image.data,vb=False )
        self.assertGreaterEqual( np.nanmin(Xhc), 0.0 )
        self.assertLessEqual( np.nanmax(Xhc), 1.0 )
        self.assertEqual( image.data.shape, Xhc.shape )

        # test hull correction on HyData instances
        for D in [image, cloud]:
            Xhc = get_hull_corrected(D,vb=False)
            self.assertEqual(image.data.shape, Xhc.data.shape)
            self.assertGreaterEqual(np.nanmin(Xhc.data), 0.0)
            self.assertLessEqual(np.nanmax(Xhc.data), 1.0)

    def test_mwl(self):
        image = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))
        image.data[:50,:,:] = np.nan # add some nans to make more realistic
        cloud = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))

        # also create a HyLibrary instance
        image.header.set_sample_points('A', [(20, 15)]) # label some seed pixels in each sample
        image.header.set_sample_points('B', [(80, 15)])
        image.header.set_sample_points('C', [(140, 15)])
        image.header['class names'] = ['A', 'B', 'C']
        from hylite.hylibrary import from_indices
        lib = from_indices(image,
                           [image.header.get_sample_points(n)[0] for n in image.header.get_class_names()],
                           names=image.header.get_class_names(),
                           s=5)


        for D in [lib,image,cloud]:
            # test normal mwl
            mwl = minimum_wavelength(D, 2100., 2380.,
                                     trend='hull', method='minmax',
                                     n=1, nthreads=1, vb=False, xtol=0.1, ftol=0.1)
            self.assertGreater(np.nanmax(mwl.model.data[..., 1]), 2100.)  # check some valid features were identified

            mwl = minimum_wavelength(D, 2100., 2380., trend='hull', method='quad', n=2, nthreads=1, vb=True, xtol=0.1,
                                     ftol=0.1)

            self.assertGreater(np.nanmax(mwl.model.data[..., 1]), 2100.)  # check some valid features were identified

            mwl = minimum_wavelength(D, 2100., 2380., trend='hull', method='poly', n=2, nthreads=1, vb=True, xtol=0.1,
                                     ftol=0.1)

            self.assertGreater(np.nanmax(mwl.model.data[..., 1]), 2100.)  # check some valid features were identified

            mwl = minimum_wavelength(D, 2100., 2380., trend='hull', method='gauss', n=1, nthreads=1, vb=True, xtol=0.1, ftol=0.1)
            self.assertGreater(np.nanmax(mwl.model.data[..., 1]), 2100.)  # check some valid features were identified

            # test multi-mwl
            M = minimum_wavelength( D, minw=2100., maxw=2400., sym=False,method='gauss',n=3, vb=True, xtol=0.1, ftol=0.1)
            mask = np.isfinite(M[0, 0])

            # check depth sorting
            M.sortByDepth()
            self.assertTrue( (M[0, 'depth'][mask] >= M[1, 'depth'][mask]).all())
            self.assertTrue( (M[2, 'depth'][mask] >= M[2, 'depth'][mask]).all())

            # check pos sorting
            M.sortByPos()
            self.assertTrue( (M[0, 'pos'][mask] <= M[1, 'pos'][mask]).all())
            self.assertTrue( (M[2, 'pos'][mask] <= M[2, 'pos'][mask]).all())

            # check deepest feature extraction
            deepest = M.deepest(2100., 2400.)
            M.sortByDepth()
            self.assertTrue( (np.nan_to_num(deepest.data) == np.nan_to_num(M[0].data)).all() )

            # check closest feature extraction
            closest = M.closest(2100.)
            M.sortByPos()
            self.assertTrue( (np.nan_to_num(closest.data) == np.nan_to_num(M[0].data)).all() )

            closest = M.closest(2200., valid_range=(2195., 2205.))
            self.assertTrue( np.nanmin(closest.data[..., 1]) >= 2195., "Error - %s" % np.nanmin(closest.data[..., 1]))
            self.assertTrue( np.nanmin(closest.data[..., 1]) <= 2205., "Error - %s" % np.nanmin(closest.data[..., 1]))

            # test colourise function [ just run it to check for crashes ]
            rgb, leg = colourise_mwl( M.closest(2200., valid_range=(2150., 2230.) ) )
            leg.quick_plot()

            # test multithreading (again; just see if it runs for now )
            M1 = minimum_wavelength( D, minw=2100., maxw=2400., sym=True, method='gauss',n=3, vb=True, nthreads=-1, xtol=0.1, ftol=0.1)

            # run some other fun functions
            ev = M1.evaluate()
            cls = M1.classify(5,nf=3)
            rs = M1.residual()

            # run plotting code
            fig,ax = M1.quick_plot()
            fig, ax = M1.quick_plot(step=3)

            # fit peaks
            M = minimum_wavelength(D, minw=2100., maxw=2400., sym=False, method='gauss', n=3, vb=True, xtol=0.1,
                                   ftol=0.1, minima=False)
            M.sortByDepth()
            self.assertGreater(np.nanmax(M[0,'depth']), 0 ) # check there are some valid features

    def testIO(self):
        image = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))
        M = minimum_wavelength(image, minw=2100., maxw=2400., sym=False, method='gauss', n=2, vb=True)
        pth = mkdtemp()

        # do save and load
        eq0,eq1 = False,False
        df = np.inf
        try:
            io.save(pth+'/test', M)
            print("MWL: ", os.path.exists('/Users/thiele67/Documents/Python/public/hylite/tests/M.mwl'))
            eq0 = os.path.exists(os.path.join(pth,'test.mwl')) # save worked?
            M2 = io.load(os.path.join(pth,'test.hdr'))

            # check identical
            eq1 = (M2.x == M.x).all()
            df = np.nanmax( M.model.data - M2.model.data ) # max difference

        except:
            shutil.rmtree(pth)  # delete temp directory
            self.assertFalse(True, "Error..." )

        shutil.rmtree(pth)  # delete temp directory

        self.assertTrue(eq0) # fail here if something went wrong
        self.assertTrue(eq1)
        self.assertTrue(df < 1e-2) # difference should be very small

    def test_TPT(self):
        from hylite.filter import TPT
        image = io.load(os.path.join(os.path.join(str(Path(__file__).parent.parent), "test_data"),"image.hdr"))
        # tpt,p,d = TPT(image, sigma=10., window=7, thresh=0, vb=False)

if __name__ == '__main__':
    unittest.main()
